open Misc
open Exp
open Symbtree

exception Abstraction_error of string

let bogus_id = Misc.create_ident "_bogus"
let bogus_ids = ref []

(* Option monad *)
let (>>=) x f =
  match x with
  | None -> None
  | Some v -> f v

(** number of calls to the theorem prover during abstraction *)
let prover_calls = ref 0

(** {2 Abstraction of sequences and sets} *)

(** Get a list of all `interesting' expressions in [sl] *)
let get_values sfun sl = 
  let exp_to_list e = match e with 
    | Ce_fun(s,el) when Pervasives.(=) s sfun -> el 
    | e -> [e] in 
  (** Get all the list expressions *)
  let get_lval res s = match s with
    | Csp_node(_,_,cel) ->
	let rec go_cel = function 
	  | [] -> res 
	  | ("val",e)::_ -> [mk_fun("@item",[e])]::(exp_to_list e)::res
	  | _::cel -> go_cel cel in
	go_cel (fld_to_list cel)
    | Csp_listsegi(SINGLE _,_,_,e,_,_) -> (exp_to_list e)::res
    | _ -> res in
  List.fold_left get_lval [] sl

(** Apply a sequence substitution *)
let list_sub_can_exp sub e =
  let all = List.sort (fun (x,_) (y,_) -> compare (List.length x) (List.length y)) sub in 
  let rec go m mem = function
    | [] -> List.rev mem
    | x::xs -> 
	let m = List.filter (function (z::_,_) -> equal_exp z x | _ -> false) m in
	let m = List.map (fun (x,y) -> (List.tl x, y)) m in
	match m with 
	  | [] -> List.rev_append mem (x :: go all [] xs)
	  | ([],y)::_ -> y :: go all [] xs
	  | _ -> go m (x::mem) xs in
  let rec go_exp = function
    | Ce_fun ("@list",el) -> mk_fun ("@list", go all [] el)
    | e -> e in
  go_exp e

(** Apply a sequence substitution *)
let set_sub_can_exp sub e =
  let all = List.map (fun (x,y) -> (List.sort compare_exp x, y)) sub in
  let rec match_one = function
    | ([], ys) -> Some ys
    | (_::_, []) -> None
    | (x::xs, y::ys) ->
	if equal_exp x y then match_one (xs,ys)
	else match_one (x::xs,ys) >>= fun res -> Some (y::res) in
  let go el (x,y) = match match_one (x, el) with 
    | None -> el
    | Some el' -> List.merge compare_exp [y] el' in
  let go_exp = function
    | Ce_fun ("@set",el) -> 
	let el = List.sort compare_exp el in
	mk_fun ("@set", List.fold_left go el all)
    | e -> e in
  go_exp e

(** Abstract list values *)
let abs_list_values (pl,sl) =
  (** Break [ys] into two sublists if it contains [x]. *)
  let rec break1 x zs e ys = match ys with
    | [] -> (List.rev zs)::e
    | y::ys ->
	if equal_exp x y then
	  (List.rev zs)::ys::e
	else
	  break1 x (y::zs) e ys in
  let rec break2 x zs e ys = match ys with
    | [] -> (List.rev zs)::e
    | y::ys ->
	if equal_exp x y then
	  (List.rev zs)::(y::ys)::e
	else
	  break2 x (y::zs) e ys in
  let rec break3 zs e ys = match ys with
    | [] -> (List.rev zs)::e
    | Ce_fun("@item",[y])::ys when IdSet.is_empty (ex_fv_can_exp y) ->
	break3 [] ((List.rev zs)::e) ys
    | y::ys ->
	break3 (y::zs) e ys in
  let rec reduce n xs = 
    if n > List.length xs then xs 
    else match xs with
      | [] -> [] 
      | []::xs -> reduce n xs
      | [y]::xs ->
	 let xs = List.fold_left (break1 y []) [] (List.rev xs) in
	 reduce n xs
      | (y::ys)::xs -> 
	 let xs = List.fold_left (break2 y []) [y::ys] (List.rev xs) in
	 reduce (n+1) xs in
  let xs = get_values "@list" sl in
(*  let _ = Format.printf "List abstraction %a@." pp_can_form ((pl,sl),[]) in 
  let _ = Format.printf "List abstraction %d %d@." (List.length xs)
	(List.fold_left (fun e x -> e + List.length x) 0 xs)  in 
  List.iter (fun el -> Format.printf "____ %a@." pp_can_exp (mk_fun("@list", el))) xs; *)
  let xs = List.fold_left (break3 []) [] (List.rev xs) in
  let xs = reduce 0 xs in
  let xs = List.map List.rev xs in
  let xs = reduce 0 xs in
  let xs = List.map List.rev xs in
(*  let _ = Format.printf "List abs_2       %d %d@." (List.length xs)
	(List.fold_left (fun e x -> e + List.length x) 0 xs)  in
  List.iter (fun el -> Format.printf "____ %a@." pp_can_exp (mk_fun("@list", el))) xs; *)
  let xxs = List.map (fun x -> 
    let e = Ce_ident (gensym (create_ident "_VAL")) in
    let e = match x with 
      | [Ce_fun("@item",_)] -> mk_fun("@item", [e])
      | _ -> e in
    (x, e)) xs in
  match map_uform (list_sub_can_exp xxs) (pl,sl) with
  | Some uf -> (*Format.printf "List abs result  %a@." pp_can_form (uf,[]);*)
	       uf
  | None -> assert false

(** Abstract set values *)
let abs_set_values (pl,sl) =
  (** Break [ys] into two sublists if it contains [x]. *)
  let rec break xs e ys = 
    match List.partition (fun x -> mem_exp x xs) ys with
      | ([], ys) -> ys::e
      | (ys, []) -> ys::e
      | (y1, y2) -> y1::y2::e in
  let rec reduce n xs = 
    if n > List.length xs then xs 
    else match xs with
      | [] -> [] 
      | []::xs -> reduce n xs
      | [x]::xs ->
	 let xs = List.map (List.filter (fun y -> not (equal_exp x y))) xs in
	 reduce n xs
      | ys::xs -> 
	 let xs = List.fold_left (break ys) [ys] (List.rev xs) in
	 reduce (n+1) xs in
  let xs = get_values "@set" sl in
(*  let _ = Format.printf "Set abstraction %a@." pp_can_form ((pl,sl),[]) in 
  let _ = Format.printf "Set abstraction %d %d@." (List.length xs)
	(List.fold_left (fun e x -> e + List.length x) 0 xs)  in 
  List.iter (fun el -> Format.printf "____ %a@." pp_can_exp (mk_fun("@set", el))) xs;  *)
  let xs = reduce 0 xs in
(*  let _ = Format.printf "Set abs_2       %d %d@." (List.length xs)
	(List.fold_left (fun e x -> e + List.length x) 0 xs)  in
  List.iter (fun el -> Format.printf "____ %a@." pp_can_exp (mk_fun("@set", el))) xs;  *)
  let xxs = List.map (fun x -> 
    let e = Ce_ident (gensym (create_ident "_VAL")) in
    let e = match x with 
      | [Ce_fun("@item",_)] -> mk_fun("@item", [e])
      | _ -> e in
    (x, e)) xs in
  match map_uform (set_sub_can_exp xxs) (pl,sl) with
  | Some uf -> (*Format.printf "Set abs result  %a@." pp_can_form (uf,[]); *)
	       uf
  | None -> assert false


(* Convert to a normal form *)
let abs_fv_to_sub exfv = 
  let rec combine xl yl = match xl,yl with
    | [],_ -> []
    | x::xl,[] -> raise (Invalid_argument "ufl_to_nf combine")
    | x::xl,y::yl -> (x,y)::combine xl yl in
  while (List.length !bogus_ids <  List.length exfv) do
    bogus_ids := !bogus_ids @ (Ce_ident (Misc.gensym bogus_id) :: 
                    List.map (fun _ -> Ce_ident (Misc.gensym bogus_id)) !bogus_ids)
  done;
  mk_subst (combine exfv !bogus_ids)


let uf_to_nf (pl,sl) =
  match aggr_norm_pure pl with None -> None | Some pl ->
(*TODO  let pl = List.sort (compare_atom ~cmp_id:Misc.abs_compare_ident) pl in *)
  let sl = List.sort (compare_can_spred ~cmp_id:Misc.abs_compare_ident) sl in
  let exfv = abs_fv_can_uform (pl,sl) in
  let sub = abs_fv_to_sub exfv in
  match map_uform sub (pl,sl) with None -> assert false | Some (pl,sl) ->
(*TODO  let pl = List.sort compare_atom pl in *)
  let sl = List.sort compare_can_spred sl in
  Some (pl,sl)

let cf_to_nf (uf,scpl) =
  let exfv = abs_fv_can_uform uf in
  let uf = match map_uform (abs_fv_to_sub exfv) uf with None -> assert false | Some uf -> uf in
  let rec go (uf',scpl') =
    let exfv' = abs_fv_can_uform uf' in
    let exfv' = exfv @ List.filter (fun x -> not (List.exists (fun y -> equal_ident x y) exfv)) exfv' in
    let sub = abs_fv_to_sub exfv' in
    match map_form sub (uf',scpl') with None -> assert false | Some ((pl',sl'),scpl') ->
(*TODO    let pl' = List.sort compare_atom pl' in *)
    let sl' = List.sort compare_can_spred sl' in
    ((pl',sl'),scpl') in
  let scpl = List.map (fun (s,cp) -> (s, {cfl = List.sort compare_can_form (List.map go cp.cfl)})) scpl in
  (uf,scpl)

(* Abstraction *)

let link_of_node snode =
  try 
    assert (Pervasives.(=) (String.get snode 0) '.');
    String.sub snode 1 ((String.length snode)- 1)
  with Not_found -> assert false

(* Using the assumptions pl, if s is a singly-linked list segment
   with link field t and value field v, return
    (0) the type of node
    (1) the connection function
    (2) the end of the list segment 
    (3) the value of the list segment
    (4) the set of expressions not contained in it
    (5) the list of other fields
 *)
let get_ends pl t v s = match s with
  | Csp_node(nid,e,fields) -> if Pervasives.(=) t "val" then None else 
      (try
	let f = fld_get t fields in
	let fv = 
	  try fld_get v fields
          with Not_found -> Ce_ident (gensym_garb (create_ident "VAL")) in
        let fv = mk_fun ("@item", [fv]) in
	let fields = fld_remove t fields in
	let fields = fld_filter (function (t,Ce_num _)->true| _ -> false) fields in
	Some (nid,"@item",f,fv,get_uneql e pl,fields)
      with Not_found ->
        None)
  | Csp_listsegi(SINGLE(snode,sfun,flds),_,f1,e2,_,el) when Pervasives.(=) (link_of_node snode) t ->
      Some ("Node",sfun,f1,e2,el,flds)
  | _ -> None (*raise (Abstraction_error "ABS: Different field found")*)

(* Using the assumptions pl, if s is a singly-linked list segment
   ending at e0, return
    (0) the type of node
    (1) the connection function
    (2) its start
    (3) its link field
    (4) its value
    (5) the list of other fields 
    (6) the set of expressions not contained in it
 *)
let get_ends_with pl e0 s = match s with
  | Csp_node(nid,e,fields) -> 
      let t = fld_containing e0 fields in
      let fv = 
	try fld_get "val" fields
        with Not_found -> Ce_ident (gensym_garb (create_ident "VAL")) in
      let fv = mk_fun ("@item", [fv]) in
      let fields = fld_remove t fields in
      let fields = fld_filter (function (t,Ce_num _)->true| _ -> false) fields in
      (nid,"@item",e,t,fv,fields,get_uneql e pl)
  | Csp_listsegi(SINGLE(snode,sfun,fields),e1,f1,e2,_,el) when equal_exp f1 e0 ->
      ("Node",sfun,e1,link_of_node snode,e2,fields,el)
  | _ -> raise (Abstraction_error "ABS: Different field found")

let normal_uform prover uf  =
  match prover.normal_can_form (uf,[]) with
    | None -> None
    | Some (uf',[]) -> Some uf'
    | _ -> assert false

let combine_funs (sfun1,fv) (sfun2,fv') =
  match (sfun1, sfun2) with
    | ("@zero", _) -> ("@zero",Ce_num 0)
    | (_, "@zero") -> ("@zero",Ce_num 0)
    | ("@item", "@item") -> (!Config.value_abstraction, mk_fun(!Config.value_abstraction,[fv;fv']))
    | ("@item", _) -> (sfun2, mk_fun(sfun2,[fv;fv']))
    | (_, "@item") -> (sfun1, mk_fun(sfun1,[fv;fv']))
    | _ -> 
	assert (Pervasives.(=) sfun1 sfun2);
	(sfun1, mk_fun(sfun1,[fv;fv']))

(* Applies the abstraction rules to [uf] to eliminate some existentials *)
let abs_can_uform prover uf =
  let start_of s = match s with
    | Csp_node(_,e,_) -> e
    | Csp_listsegi(SINGLE _,e,_,_,_,_) -> e
    | Csp_listsegi(_,_,_,_,_,_) -> raise  (Abstraction_error "ABS: Cannot use dllseg or xlseg")
    | Csp_tree _ -> raise (Abstraction_error "ABS: Cannot use tree")
    | Csp_indpred _ -> raise (Abstraction_error "ABS: Cannot use inductive predicate") in
  let ends_with_exp e0 s = match s with
    | Csp_node(_,_,fields) -> fld_exists (function (_,e)-> equal_exp e e0) fields
    | Csp_listsegi(SINGLE _,_,e,_,_,_) -> equal_exp e e0
    | Csp_listsegi(_,_,_,_,_,_) -> raise (Abstraction_error "ABS: Cannot use dllseg or xlseg")
    | Csp_tree _ -> raise (Abstraction_error "ABS: Cannot use tree")
    | Csp_indpred _ -> raise (Abstraction_error "ABS: Cannot use inductive predicate") in
  let post_proc changed pl t v s = match s with
(*    | Csp_node (_,e,_) -> 
	begin match get_ends pl t v s with
	  | None -> (changed, s)
	  | Some (nid,sfun,f,fv,el',flds) ->
	      let fv = mk_fun(!Config.value_abstraction,[fv]) in
	      (true, Csp_listsegi(SINGLE("."^t,!Config.value_abstraction,flds),e,f,fv,Ce_num 0,el'))
	end *)
    | _ -> (changed, s) in
  let rec go changed pl (sl1,sl2) = match sl1 with
    | [] -> if changed then go false pl (sl2,[]) else sl2
    | s::sl1 ->
	let e_mid = start_of s in
	if not (existential_id e_mid) then
	  go changed pl (sl1,s::sl2) 
	else
	  let (sl1a,sl1b) = List.partition (ends_with_exp e_mid) sl1 in
	  let (sl2a,sl2b) = List.partition (ends_with_exp e_mid) sl2 in
	  begin match (sl1a@sl2a) with
	    | [] -> if !Config.allow_leaks then go changed pl (sl1,sl2) else raise Junk
	    | [s'] ->
	      let (nid1,sfun1,e,t,fv1,flds,el) = get_ends_with pl e_mid s' in
	      begin match get_ends pl t "val" s with
		| None -> go changed pl(sl1,s::sl2)
		| Some (nid2,sfun2,f,fv2,el',flds') ->
		    if Pervasives.(=) nid1 nid2 then begin
		      (*let (sfun2,f,fv2,el',flds') = get_ends pl t "val" s in*)
		      let (sfun,fv) = combine_funs (sfun1,fv1) (sfun2,fv2) in
		      let el'' = dset_inter el el' in
		      let flds'' = fld_inter flds flds' in
		      go true pl (Csp_listsegi(SINGLE("."^t,sfun,flds''),e,f,fv,Ce_num 0,el'')::sl1b,sl2b)
		    end else
		      go changed pl (sl1,s::sl2)
	      end
	    | s'::_::_ ->
	      let (_,_,_,t,_,_,_) = get_ends_with pl e_mid s' in
	      let (changed,s) = post_proc changed pl t "val" s in
	      go changed pl (sl1,s::sl2)
	  end in
  let first_pass (sub,sl) = function
    | Csp_node(nid,e,cel) ->
	let go_fld (sub,cel) (c,e) =
	  if Config.is_lock_field c then
	    let (sub,e) = match e with
	      | Ce_num x -> (sub,e)
	      | Ce_ident id ->
		  if equal_ident id tid_var then (sub, e) else
		  let idn = Ce_ident (gensym (create_ident "_LOCK")) in
		  ((id,idn)::sub, idn)
	      | _ -> 
		  let idn = Ce_ident (gensym (create_ident "_LOCK")) in
		  (sub, idn) in
            (sub,(c,e)::cel)
	  else (sub,(c,e)::cel) in
	let (sub,cel) = List.fold_left go_fld (sub,[]) (fld_to_list cel) in
	(sub, Csp_node(nid,e,fld_from_list cel)::sl)
    | sp -> (sub,sp::sl) in
  normal_uform prover uf >>= fun (pl,sl) ->
  let _ = if !Config.verbose2 then Format.printf "ABS %a@." pp_can_form ((pl,sl),[]) in
  let (sub, sl) = List.fold_left first_pass ([],[]) sl in
  map_uform (mk_subst sub) (pl,sl) >>= fun (pl,sl) ->
  let sl = go false pl (sl,[]) in
  (match kill_garbage_vars (empty_eq,{cfl = [((pl,sl),[])]}) with
    | _, {cfl = [(uf,_)]} ->
        normal_uform prover uf >>= fun uf ->
        let uf =
	  if Pervasives.(=) !Config.value_abstraction "@list" then abs_list_values uf 
	  else if Pervasives.(=) !Config.value_abstraction "@set" then abs_set_values uf 
	  else uf in
        normal_uform prover uf 
    | _ -> assert false)

let abs_can_form prover (uf,scpl) = abs_can_uform prover uf >>= fun uf -> Some (uf,scpl)

let implies prover cf1 cf2 = 
  let exfv = ex_fv_can_prop {cfl = [cf1]} in
  let sub = existentials_rename_tonormal_sub exfv in
  match map_form sub cf1 with
  | None ->  assert false 
  | Some cf1' -> incr prover_calls; prover.entails_form cf1' cf2

let implies_uf prover uf1 uf2 =
  implies prover (uf1,[]) (uf2,[])

(** Normalize [ufl] *)
let uform_list_simplify prover ufl =
  let ufl = Misc.map_partial (normal_uform prover) ufl in 
  let ufl = Misc.map_partial uf_to_nf ufl in
  let ufl = List.sort compare_uform ufl in 
  let ufl = remove_uform_duplicates ufl in
  ufl

let cform_list_simplify prover cfl =
  let cfl = Misc.map_partial (prover.normal_can_form) cfl in
  let cfl = List.map cf_to_nf cfl in
  let cfl = List.sort compare_can_form cfl in 
  let cfl = remove_can_form_duplicates cfl in
  cfl

(** The function [can_prop_to_uform_list cp] converts a can_prop [cp] into a can_uform list.
    It assumes that [cp] contains no boxes.  *)
let can_prop_to_uform_list prover {cfl=cfl} = 
  let ufl = List.map (fun (uf,scpl) -> assert(Pervasives.(=) scpl []); uf) cfl in
  uform_list_simplify prover ufl

let can_prop_to_can_form_list prover {cfl=cfl} =
  cform_list_simplify prover cfl

(** The function [cform_list_to_can_prop cfl] converts a can_form list [cfl] into a can_prop. *)
let can_form_list_to_can_prop cfl = {cfl=cfl} 

(** The function [uform_list_to_can_prop ufl] converts a can_uform list [ufl] into a can_prop. *)
let uform_list_to_can_prop ufl =
  can_form_list_to_can_prop (List.map (fun x -> (x,[])) ufl)

let adds_info prover cfl cf = not (List.exists (fun cf1 -> compare_can_form cf cf1 = 0 || implies prover cf cf1) cfl)

(** Test whether [uf] adds any new information as a disjunct to [ufl] *)
let adds_info_uf prover ufl uf = not (List.exists (fun uf1 -> compare_uform uf uf1 = 0 || implies_uf prover uf uf1) ufl)


let match_exp e1 e2 =
  if equal_exp e1 e2 then Some (e1,pure_true,pure_true) else
  match (e1,e2) with
  | (Ce_num n1, Ce_num n2) ->
      if n1=n2 then
	Some (Ce_num n1, pure_true, pure_true)
      else None
  | (Ce_ident id, e2) when Misc.is_existential id ->
      Some(Ce_ident id, pure_true, pure_one (mk_EQ(Ce_ident id, e2)))
  | (e1, Ce_ident id) when Misc.is_existential id ->
      Some(Ce_ident id, pure_one (mk_EQ(Ce_ident id, e1)), pure_true)
  | (e1, e2) ->
      let id = gensym (create_ident "_match") in
      Some(Ce_ident id, pure_one (mk_EQ(Ce_ident id, e1)), pure_one (mk_EQ(Ce_ident id, e2)))

let match_fields (pl1,pl2) (fld1,fld2) = 
  let rec go (cel,pl1,pl2) = function
  | ([], []) -> Some (cel,pl1,pl2)
  | ([], (x2,e2)::cel2) -> 
      let e = Ce_ident (gensym (create_ident "_match")) in
      go ((x2,e)::cel, pl1, pure_conj_one (mk_EQ(e,e2)) pl2) ([],cel2)
  | ((x1,e1)::cel1, []) -> 
      let e = Ce_ident (gensym (create_ident "_match")) in
      go ((x1,e)::cel, pure_conj_one (mk_EQ(e,e1)) pl1, pl2) (cel1,[])
  | ((x1,e1)::cel1, (x2,e2)::cel2) when Pervasives.(=) x1 x2 -> (* Main case *)
      match_exp e1 e2 >>= fun (e,pl1',pl2') ->
      go ((x1,e)::cel, pure_conj pl1' pl1, pure_conj pl2' pl2) (cel1,cel2)
  | ((x1,e1)::cel1, ((x2,_)::_ as cel2)) when x1<x2 -> 
      let e = Ce_ident (gensym (create_ident "_match")) in
      go ((x1,e)::cel, pure_conj_one (mk_EQ(e,e1)) pl1, pl2) (cel1,cel2)
  | (((x1,_)::_ as cel1), (x2,e2)::cel2) (* when x1>x2 *) -> 
      let e = Ce_ident (gensym (create_ident "_match")) in
      go ((x2,e)::cel, pl1, pure_conj_one (mk_EQ(e,e2)) pl2) (cel1,cel2)
  in
  go ([],pl1,pl2) (fld_to_list fld1, fld_to_list fld2) >>= fun (cel,pl1,pl2) ->
  aggr_norm_pure pl1 >>= fun pl1 ->
  aggr_norm_pure pl2 >>= fun pl2 ->
  Some (fld_from_list cel, pl1, pl2)


let match_spred (pl1,pl2) = function
  | (Csp_listsegi(SINGLE(snode1,sfun1,fld1),e1,f1,g1,_,ds1),
     Csp_listsegi(SINGLE(snode2,sfun2,fld2),e2,f2,g2,_,ds2))
    when Pervasives.(=) snode1 snode2 && Pervasives.(=) sfun1 sfun2 && compare_fields fld1 fld2 = 0 ->
      if equal_exp e1 e2 && equal_exp f1 f2 && equal_exp g1 g2 (*TODO*) then
	let ds = dset_inter ds1 ds2 in
	Some (Csp_listsegi(SINGLE(snode1,sfun1,fld1),e1,f1,g1,Ce_num 0,ds),pl1,pl2)
      else
	None
  | (Csp_node(nid1,e1,fld1), Csp_node(nid2,e2,fld2)) 
    when Pervasives.(=) nid1 nid2 && equal_exp e1 e2 ->
      match_fields (pl1,pl2) (fld1,fld2) >>= fun (fld,pl1,pl2) ->
      Some (Csp_node(nid1,e1,fld),pl1,pl2)
  | _ -> None

let match_uform (pl1,sl1) (pl2,sl2) =
  let rec go res (pl1,pl2) = function
    | ([],[]) ->
      let pl = pure_filter (fun ca -> mem_atom ca pl1) pl2 in
      Some (pl, List.rev res)
    | ([], _::_) -> None
    | (_::_, []) -> None
    | (s1::sl1, s2::sl2) ->
      match_spred (pl1,pl2) (s1,s2) >>= fun (sp,pl1,pl2) ->
      go (sp::res) (pl1,pl2) (sl1,sl2)
  in
    go [] (pl1,pl2) (sl1,sl2)

let combine_ufl prover (ufl_old,ufl_new) =
  let f uf1 uf2 = match match_uform uf1 uf2 with
    | None -> uf2
    | Some uf -> uf in
  let rec go n (ufl_done,ufl_old,ufl,nufl) = match nufl with
    | [] -> ufl@ufl_done
    | (n1,pl,sl)::nufl when n = n1 && n1 mod 2=0 -> (* ufl_old *)
	go n (ufl_done,(pl,sl)::ufl_old,ufl,nufl)
    | (n1,pl,sl)::nufl when n/2 = n1/2 -> (* ufl_new *)
	let ufl = map_partial (match_uform (pl,sl)) ufl_old @ List.map (f (pl,sl)) ufl in
	let ufl = if adds_info_uf prover ufl (pl,sl) then (pl,sl)::ufl else ufl in
	go n (ufl_done,ufl_old,ufl,nufl)
    | (n1,_,_)::_ (* when n<n1 *) -> go n1 (ufl@ufl_done,[],[],nufl) in
  let nufl_old = List.map (fun (pl,sl) -> (2 * List.length sl, pl, sl)) ufl_old in
  let nufl_new = List.map (fun (pl,sl) -> (2 * List.length sl + 1, pl, sl)) ufl_new in
  let nufl = List.sort (fun (n1,_,_) (n2,_,_) -> compare n1 n2) (nufl_new @ nufl_old) in
  go 0 ([],[],[],nufl)

(*
let aggr_simp_uform_list prover very_aggressive ufl = 
  let rec rem_disj (ufl1,ufl2) = match ufl1 with
    | [] -> ufl2 
    | uf::ufl1 -> 
	let ufl2' = List.filter (fun uf2 -> not (implies_uf prover uf2 uf)) ufl2 in
	rem_disj(ufl1,uf::ufl2') in
  let ufl =
    if not very_aggressive then ufl else
      let intersection el1 el2 = List.filter (fun x -> mem_atom x el2) el1 in
      let mk_pair p = if is_EQ p then un_EQ p else if is_NEQ p then un_NEQ p else (Ce_num 0,Ce_num 0) in
      let f r (pl,sl) = List.filter (fun p -> mem_atom (mk_EQ p) pl ||  mem_atom (mk_NEQ p) pl) r in
      (* remove some inequalities *)
      let e2l = match ufl with
                | [] -> []
                | (pl,_)::ufl -> List.fold_left f (List.map mk_pair pl) ufl in
      let pl' = List.map mk_NEQ e2l in
      List.map (fun (pl,sl) -> (List.filter is_EQ pl @ intersection pl' pl, sl)) ufl in
  (* test implications *)
  let ufl = rem_disj (ufl,[]) in (* testing implications one way round *)
  let ufl = rem_disj (ufl,[]) in (* and again in reverse *)
  ufl
*)

(** atempt to merge two disjuncts into one *)
let try_merge prover ((pl1,sl1),scpl1) ((pl2,sl2),scpl2) = 
  let go filter (s,cp) = (s,{cfl=List.map (fun ((pl,sl),_) -> ((pure_filter filter pl,sl),[])) cp.cfl}) in
  let pl = pure_filter (fun ca -> mem_atom ca pl2) pl1 in
  let pl1a = pure_filter (fun ca -> not (mem_atom ca pl2)) pl1 in
  let pl2a = pure_filter (fun ca -> not (mem_atom ca pl)) pl2 in
  let scpl1a = List.map (go (fun ca -> not (mem_atom ca pl1a))) scpl1 in
  let scpl2a = List.map (go (fun ca -> not (mem_atom ca pl2a))) scpl2 in
  if implies prover ((pl1,sl1),scpl1) ((pl2,sl2),scpl2) then
    Some ((pl2,sl2),scpl2)
  else if implies prover ((pl2,sl2),scpl2) ((pl1,sl1),scpl1) then
    Some ((pl1,sl1),scpl1)
  else if implies prover ((pl1,sl1),scpl1) ((pl,sl2),scpl2a) then
    Some ((pl,sl2),scpl2a)
  else if implies prover ((pl2,sl2),scpl2) ((pl,sl1),scpl1a) then
    Some ((pl,sl1),scpl1a)
  else
    None

let aggr_simp_can_form_list prover cfl =
  let rec do_merges f (cfl_old,cfl_new) cfl = match cfl with
    | [] -> (cfl_old, cfl_new)
    | cf::cfl -> match f cf with 
       | None -> do_merges f (cf::cfl_old, cfl_new) cfl
       | Some cf -> do_merges f (cfl_old, cf::cfl_new) cfl in
  let rec rem_disj (cfl1,cfl2) = match cfl1 with
    | [] -> cfl2 
    | cf::cfl1 -> 
	let (cfl2old,cfl2new) = do_merges (try_merge prover cf) ([],[]) cfl2 in
	match cfl2new with 
	| [] -> rem_disj (cfl1, cf::cfl2old)
        | _::_ -> rem_disj (cfl2new@cfl1, cfl2old) in
  rem_disj (cfl,[])


(** split uform list *)
let split_uform_list filter ufl =  
  let split_spatial (pl,sl) =
    let (sl1,sl2) = List.partition (fun x -> filter x <> None) sl
    in match sl1 with
       | s::sl1' -> (filter s,pl,sl1'@sl2)
       | [] -> (None,pl,sl2) in
  let rec go (r1,r2) xl = match xl with
    | [] -> (r1,r2)
    | (Some x,pl,sl)::xl -> go ((x,(pl,sl))::r1,r2) xl
    | (None,pl,sl)::xl -> go (r1,(pl,sl)::r2) xl in
  go ([],[]) (List.rev_map split_spatial ufl)

let ptsto_filter e = function
  | Csp_node(_,e',el') when equal_exp e e' -> Some el'
  | _ -> None

let compare_fields_uform (x1,y1) (x2,y2) =
  let n = compare_fields x1 x2 in if n<>0 then n else 
  compare_uform y1 y2

let mem_fields_uform y l =
  List.exists (fun x -> compare_fields_uform x y = 0) l

(* Turn  (e|->x:e1 * p) || (e|->x:e2 * p)  into (e|-> * p)  *)
let forget_field_values e (ufl_old,ufl) =
  let compare_snd x y = compare_uform (snd x) (snd y) in
  let rec go xufl = match xufl with
    | [] -> []
    | [xuf] -> [xuf]
    | (x1,uf1)::(x2,uf2)::xufl ->
        if compare_uform uf1 uf2 = 0 (*&& (existential_id e || false)*) then
	  let _ = if !Config.verbose2 then
	     Format.fprintf (!Config.formatter) "forgot field %a@." pp_can_exp e in
	  go ((fld_inter x1 x2,uf1)::xufl)
	else
	  (x1,uf1)::go ((x2,uf2)::xufl) in
  let (xufl,ufl) = split_uform_list (ptsto_filter e) ufl in
  let (xufl_old,_) = split_uform_list (ptsto_filter e) ufl_old in
  let xufl = List.sort compare_snd (xufl@xufl_old) in
  let xufl = go xufl in
  let xufl = List.filter (fun xuf -> not (mem_fields_uform xuf xufl_old)) xufl in
  let ufl' = List.map (fun(se,(pl,sl)) -> (pl,List.merge compare_can_spred [Csp_node("Node",e,se)] sl)) xufl in
  ufl'@ufl

module ExpSet =
  Set.Make(struct
    type t = can_exp
    let compare = fun x -> compare_exp x
  end)

let forget_fields (ufl_old,ufl) =
  let esr = ref ExpSet.empty in
  let go = function 
   | Csp_node(_,e,_) -> esr := ExpSet.add e !esr
   | _ -> () in
  List.iter (fun(_,sl)->List.iter go sl) ufl;
  List.iter (fun(_,sl)->List.iter go sl) ufl_old;
  let uflr = ref ufl in
  ExpSet.iter (fun e -> uflr := forget_field_values e (ufl_old,!uflr)) !esr;
  !uflr


let uform_list_abs prover ufl =
  let ufl = Misc.map_partial (abs_can_uform prover) ufl in
  let ufl = uform_list_simplify prover ufl in
  ufl

let uform_list_join prover ufl_total ufl ufl_new =
  let rec rem_disj (ufl1,ufl2) = match ufl1 with
    | [] -> ufl2 
    | uf::ufl1 -> 
	let ufl2' = List.filter (fun uf2 -> not (implies_uf prover uf2 uf)) ufl2 in
	rem_disj(ufl1,uf::ufl2') in
  let ufl_new = uform_list_abs prover ufl_new in
  let ufl_new = List.filter (adds_info_uf prover ufl_total) ufl_new in
  let ufl_new = (* -- Can still be problematic!!! *)
    if !Config.forget_fields then uform_list_simplify prover (combine_ufl prover (ufl_total,ufl_new))
    else ufl_new in 
  let ufl_total = ufl_new @ ufl_total in
  let ufl = ufl_new @ ufl in
  let ufl = remove_uform_duplicates ufl in
  (* test implications *)
  let ufl = rem_disj (ufl,[]) in (* testing implications one way round *)
  let ufl = rem_disj (ufl,[]) in (* and again in reverse *)
  let ufl_new = List.filter (fun x -> List.mem x ufl) ufl_new in
  (ufl_total, ufl, ufl_new)


let prop_abs prover cp =
  let cfl = cp.cfl in
  let cfl = Misc.map_partial (abs_can_form prover) cfl in
  let cfl = cform_list_simplify prover cfl in
  {cfl=cfl}

let prop_join prover cp_total cp cp_new =
  let rec rem_disj (cfl1,cfl2) = match cfl1 with
    | [] -> cfl2 
    | cf::cfl1 -> 
	let cfl2' = List.filter (fun cf2 -> not (implies prover cf2 cf)) cfl2 in
	rem_disj(cfl1,cf::cfl2') in
  let cp_new = prop_abs prover cp_new in
  let cfl_new = List.filter (adds_info prover cp_total.cfl) cp_new.cfl in
  (*let cfl_new = aggr_simp_can_form_list prover cfl_new in*)
  let cp_total = {cfl = cfl_new @ cp_total.cfl} in
  let cfl = cfl_new @ cp.cfl in
  (* test implications *)
  let cfl = rem_disj (cfl,[]) in (* testing implications one way round *)
  let cfl = rem_disj (cfl,[]) in (* and again in reverse *)
(*  let cfl = aggr_simp_can_form_list prover cfl in *)
(*  let cfl_new = List.filter (fun x -> List.mem x cfl) cfl_new in *)
  (cp_total, {cfl=cfl}, {cfl=cfl_new})


let mk_abstraction prover =
  {
    uform_abs  = uform_list_abs prover;
    uform_join = uform_list_join prover;
    prop_abs  = prop_abs prover;
    prop_join = prop_join prover;
  }

