open Misc

(** Only allow comparisons between integers to catch uses of builtin equality.
  Use Pervasives.(=) on base types only *)
let (=) (x:int) (y:int) = x=y

type component = string

(* canonical expression *)
type can_exp =
  | Ce_num of int
  | Ce_ident of ident
  | Ce_fun of string * can_exp list
  | Ce_grp of string * int * (int * can_exp) list (* abelian group *)
(* do not create Ce_fun's directly, use mk_fun function. *)
(* do not create Ce_grp's directly, use mk_grp function. *)

(* canonical atom: e1==e2 or e1!=e2 or e1<=e2 *)
type can_atom =
  | EQ of can_exp * can_exp
  | NEQ of can_exp * can_exp
  | POS of int * (int * can_exp) list (* Ce_grp("+",n,nel) >= 0 *)

type disj_set = can_exp list
type fields = (component * can_exp) list

(* Option monad *)
let (>>=) x f =
  match x with
  | None -> None
  | Some v -> f v


(* misc ops on data structs *)
let rec equal_exp e1 e2 = if e1==e2 then true else match e1 with
 | Ce_num n1 -> (match e2 with Ce_num n2 -> n1=n2 | _ -> false)
 | Ce_ident i1 -> (match e2 with Ce_ident i2 -> equal_ident i1 i2 | _ -> false)
 | Ce_fun (s1,el1) ->
    (match e2 with Ce_fun (s2,el2) -> 
        Pervasives.(=) s1 s2 && equal_exp_list el1 el2
     | _ -> false)
 | Ce_grp (s1,n1,nel1) ->
    (match e2 with Ce_grp (s2,n2,nel2) ->
	Pervasives.(=) s1 s2 && n1=n2 && equal_int_exp_list nel1 nel2
     | _ -> false)

and equal_exp_list el1 el2 = match el1,el2 with
  | [],[] -> true
  | [],_ -> false
  | _,[] -> false
  | e1::el1',e2::el2' -> equal_exp e1 e2 && equal_exp_list el1' el2'

and equal_int_exp_list el1 el2 = match el1,el2 with
  | [],[] -> true
  | [],_ -> false
  | _,[] -> false
  | (i1,e1)::el1',(i2,e2)::el2' -> i1==i2 && equal_exp e1 e2 && equal_int_exp_list el1' el2'

and compare_exp ?(cmp_id=compare_ident) e1 e2 =
  if e1==e2 then 0 else
    match e1,e2 with
      | Ce_num n1, Ce_num n2 -> compare n1 n2
      | Ce_num _, _ -> -1
      | _, Ce_num _ -> 1
      | Ce_ident i1, Ce_ident i2 -> cmp_id i1 i2
      | Ce_ident _, _ -> -1
      | _, Ce_ident _ -> 1
      | Ce_fun (s1,el1), Ce_fun (s2,el2) ->
	  let n = compare s1 s2
	  in if n<>0 then n
	    else compare_exp_list el1 el2
      | Ce_fun _, _ -> -1
      | _, Ce_fun _ -> 1
      | Ce_grp (s1,n1,nel1), Ce_grp (s2,n2,nel2) ->
	  let n = compare s1 s2
	  in if n<>0 then n
	    else let n = compare n1 n2 in
		   if n<>0 then n
		   else compare_int_exp_list nel1 nel2

and compare_exp_list ?(cmp_id=compare_ident) el1 el2 = match el1,el2 with
  | [],[] -> 0
  | [],_ -> -1
  | _,[] -> 1
  | e1::el1',e2::el2' ->
      let n = compare_exp e1 e2
      in if n<>0 then n
	else compare_exp_list el1' el2'

and compare_int_exp_list ?(cmp_id=compare_ident) (nel1 : (int * can_exp) list)  nel2 = match nel1,nel2 with
  | [],[] -> 0
  | [],_ -> -1
  | _,[] -> 1
  | (n1,e1)::el1',(n2,e2)::el2' ->
      let n = compare n1 n2
      in if n<>0 then n else
	  let n = compare_exp e1 e2
	  in if n<>0 then n
	    else compare_int_exp_list el1' el2'


let compare_dset = compare_exp_list

let rec compare_fields ?(cmp_id=compare_ident) (sel1: (component * can_exp) list) sel2 = match sel1,sel2 with
  | [],[] -> 0
  | [],_ -> -1
  | _,[] -> 1
  | (s1,e1)::el1',(s2,e2)::el2' ->
      let n = compare s1 s2
      in if n<>0 then n else
	  let n = compare_exp e1 e2
	  in if n<>0 then n
	    else compare_fields el1' el2'

let compare_atom ?(cmp_id=compare_ident) ca1 ca2 = match ca1,ca2 with
  | EQ (e1,e2), EQ(f1,f2) ->
      let n1 = compare_exp e1 f1
      in if n1<>0 then n1
	else compare_exp e2 f2
  | EQ _, _ -> -1
  | _, EQ _ -> 1
  | NEQ(e1,e2), NEQ(f1,f2) ->
      let n1 = compare_exp e1 f1
      in if n1<>0 then n1
	else compare_exp e2 f2
  | NEQ _, _ -> -1
  | _, NEQ _ -> 1
  | POS (n1,nel1), POS (n2,nel2) ->
      let n = compare n1 n2 in
      if n<>0 then n else compare_int_exp_list nel1 nel2

let equal_atom ca1 ca2 = match ca1 with
  | EQ (e1,e2) ->
     (match ca2 with EQ(f1,f2) -> equal_exp e1 f1 && equal_exp e2 f2 | _ -> false)
  | NEQ(e1,e2) ->
     (match ca2 with NEQ(f1,f2) -> equal_exp e1 f1 && equal_exp e2 f2 | _ -> false)
  | POS (n1,nel1) ->
     (match ca2 with POS(n2,nel2) -> n1==n2 && equal_int_exp_list nel1 nel2 | _ -> false)

let rec compare_component_exp ?(cmp_id=compare_ident) ((s1:component),e1) (s2,e2) =
  let n = compare s1 s2
  in if n<>0 then n
    else compare_exp e1 e2

let compare_ident_exp ?(cmp_id=compare_ident) (i1,e1) (i2,e2) =
  let n = compare_ident i1 i2
  in if n<>0 then n
    else compare_exp e1 e2

let mem_ident i l =
  List.exists (fun x -> equal_ident x i) l

let mem_exp y l =
  List.exists (fun x -> equal_exp x y) l

let mem_eq ie iel =
  List.exists (fun ie' -> compare_ident_exp ie ie' = 0) iel

let mem_component_exp y l =
  List.exists (fun x -> compare_component_exp x y = 0) l

let mk_EQ (e1,e2) =
  if compare_exp e1 e2 <= 0
  then EQ(e1,e2)
  else EQ(e2,e1)

let mk_NEQ (e1,e2) =
  if compare_exp e1 e2 <= 0
  then NEQ(e1,e2)
  else NEQ(e2,e1)

(** Assumming that el is simplified, simplify Ce_fun(s,el) *)
let mk_fun (s,el) =
  let rec concat res e = match e with
    | Ce_fun(s',el') when Pervasives.(=) s s' -> List.fold_left concat res el'
    | e -> e::res in 
  match s with
    | "!" ->
	begin match el with 
	| [Ce_num n] -> Ce_num (if n = 0 then 1 else 0)
	| [Ce_fun ("!",[e])] -> e
	| _ -> Ce_fun(s,el)
	end
    | "@zero" -> Ce_num 0
    | "@list" ->
	if Pervasives.(=) !Config.value_abstraction "@list" then 
	  match List.rev (List.fold_left concat [] el) with
	    | [e] -> e
	    | el -> Ce_fun(s, el)
	else Ce_num 0
    | "@set" ->
	if Pervasives.(=) !Config.value_abstraction "@set" then 
	  match List.rev (List.fold_left concat [] el) with
	    | [e] -> e
	    | el -> Ce_fun(s, List.sort compare_exp el)
	else Ce_num 0
    | _ -> Ce_fun (s,el)

(* Assumming that nel is simplified, simplify Ce_grp(i,n,nel) *)
(*let mk_grp (i,n,nel) =*)
let mk_grp = function
   (_,n,[]) -> Ce_num n
 | (_,0,[(1,e)]) -> e
 | (i,n,nel) -> 
  (* compute n + (m * n') *)
  let plus_times n m n' = match i with
    | "+" -> n + (m * n')
    | "^" -> if m mod 2 = 0 then n else (lxor) n n' 
    | _ -> assert false in
  let rec go n res = function (* combine nested functions *)
    | [] -> (n, res)
    | (m,Ce_num n')::nel -> go (plus_times n m n') res nel
    | (m,Ce_grp(i',n',nel'))::nel when Pervasives.(=) i' i -> 
	let nel'' = List.map (fun(n,e)->(plus_times 0 m n,e)) nel' in
	go (plus_times n m n') res (nel''@nel) 
    | ne::nel -> go n (ne::res) nel in
  let rec go2 res = function (* put same terms together *)
    | (0,e)::nel -> go2 res nel
    | (n,e)::(m,e')::nel when equal_exp e e' -> if n+m=0 then go2 res nel else go2 res ((n+m,e)::nel)
    | ne::nel -> go2 (ne::res) nel
    | [] -> res in
  let (n,nel) = go n [] nel in
  let nel = List.sort (fun (_,x) (_,y) -> -compare_exp x y) nel in
  let nel = go2 [] nel in
  let nel = match i with
    | "^" -> List.filter (fun (n,_)->n<>0) (List.map (fun (n,e) -> (n mod 2,e)) nel)
    | _ -> nel in
  match (n,nel) with 
    | (n,[]) -> Ce_num n
    | (0,[(1,e)]) -> e 
    | _ -> Ce_grp(i,n,nel)



let mk_XOR(e1,e2) = mk_grp("^",0,[(1,e1);(1,e2)])

let mk_POS(n,nel) = 
  match mk_grp("+",n,nel) with
  | Ce_num n -> POS(n,[])
  | Ce_grp("+",n,nel) -> POS(n,nel)
  | e -> POS(0,[(1,e)])

let fold_atom f ca v = match ca with
  | EQ(e1,e2) -> f e2 (f e1 v)
  | NEQ(e1,e2) -> f e2 (f e1 v)
  | POS(_,nel) -> List.fold_left (fun v (_,e) -> f e v) v nel


(* {2 Pure expressions } *)

(*module Pure = 
  Set.Make(struct 
    type t = can_atom
    let compare = compare_atom ~cmp_id:compare_ident
  end)

(** Conjunction of canonical atoms *)
type can_pure = Pure.t

let pure_true = Pure.empty
let pure_one ca  = Pure.singleton ca
let pure_conj_one ca pl = Pure.add ca pl
let pure_conj pl1 pl2 = Pure.union pl1 pl2
let pure_fold go pl e = Pure.fold go pl e
let pure_to_list pl = Pure.elements pl
let pure_is_true pl = Pure.is_empty pl
let pure_filter f pl = Pure.filter f pl
let pure_partition f pl = Pure.partition f pl
let pure_length pl = Pure.fold (fun _ x -> x+1) pl 0

let pure_first f pl = 
  Pure.fold (fun x -> function 
		  Some y -> Some y 
		| None -> if f x then Some (x, Pure.remove x pl) else None) pl None

let compare_pure pl1 pl2 = Pure.compare pl1 pl2

let mem_atom a l = Pure.mem a l

let remove_atom_duplicates l = l
*)
(*

let rec compare_pure ?(cmp_id=compare_ident) pl1 pl2 = match pl1,pl2 with
  | [],[] -> 0
  | [],_ -> -1
  | _,[] -> 1
  | ca1::pl1', ca2::pl2' ->
      let n = compare_atom ca1 ca2
      in if n<>0 then n
	else compare_pure pl1' pl2'
*)


(** Conjunction of canonical atoms *)
type can_pure = can_atom list

let pure_true = []
let pure_one ca  = [ca]
let pure_conj_one ca pl = ca::pl
let pure_conj pl1 pl2 = pl1@pl2
let pure_fold go pl e = List.fold_left (fun x y -> go y x) e pl
let pure_to_list pl = pl
let pure_is_true pl = Pervasives.(=) pl []
let pure_filter f pl = List.filter f pl
let pure_length pl = List.length pl

let pure_first f pl = match List.partition f pl with
  | ([], _) -> None 
  | (ca::pl1,pl2) -> Some (ca, pl1@pl2)

let rec compare_pure (*?(cmp_id=compare_ident)*) pl1 pl2 = match pl1,pl2 with
  | [],[] -> 0
  | [],_ -> -1
  | _,[] -> 1
  | ca1::pl1', ca2::pl2' ->
      let n = compare_atom ca1 ca2
      in if n<>0 then n
	else compare_pure pl1' pl2'

let mem_atom a l = List.exists (fun x -> equal_atom x a) l

let remove_atom_duplicates l =
  let l2 = List.stable_sort compare_atom l in
  let rec f = function
    | x::y::xs -> if equal_atom x y then f(y::xs) else x::(f(y::xs))
    | xs -> xs
  in f l2


(**** pretty printing ****)
open Format

let pp_string f s = fprintf f "%s" s

let rec pp_listsep pp s1 s2 f = function
  | [] -> fprintf f "%s" s2
  | [x] -> fprintf f "%a" pp x
  | x::l -> fprintf f "%a%s %a" pp x s1 (pp_listsep pp s1 s2) l

let rec pp_seq pp f = function
  | [] -> ()
  | [x] -> fprintf f "%a" pp x
  | x::l -> fprintf f "%a,%a" pp x (pp_seq pp) l

let rec pp_starsep pp f = function
  | [] -> fprintf f "emp"
  | [x] -> fprintf f "%a" pp x
  | x::l -> fprintf f "%a@ * %a" pp x (pp_starsep pp) l

let rec pp_can_exp f = function
  | Ce_num n -> fprintf f "%i" n
  | Ce_ident id -> fprintf f "%a" pp_ident id
  | Ce_fun (i,el) -> fprintf f "%s(%a)" i (pp_seq pp_can_exp) el
  | Ce_grp(i,0,nel) -> fprintf f "%s(%a)" i (pp_seq pp_num_exp) nel 
  | Ce_grp(i,n,nel) -> fprintf f "%s(%i,%a)" i n (pp_seq pp_num_exp) nel

and pp_num_exp f (n,e) =
    if n=1 then pp_can_exp f e
    else if n=(-1) then fprintf f "-%a" pp_can_exp e
    else fprintf f "%i(%a)" n pp_can_exp e

let pp_can_atom f ca =
  match ca with
    | EQ (e1,e2) -> fprintf f "@[%a==%a@]" pp_can_exp e1 pp_can_exp e2
    | NEQ (e1,e2) -> fprintf f "@[%a!=%a@]" pp_can_exp e1 pp_can_exp e2
    | POS(n,nel) -> fprintf f "@[%a>=0@]" pp_can_exp (Ce_grp("+",n,nel))

let pp_const f (pl,ca) =
  match pl with
    | [] -> pp_can_atom f ca
    | _::_ -> fprintf f "@[(@[%a@]@ ==> %a)@]" (pp_starsep pp_can_atom) pl pp_can_atom ca

let pp_disjoint_set f = function
  | [] -> fprintf f ""
  | el -> fprintf f ",{%a}" (pp_seq pp_can_exp) el

let pp_eq f eq =
  (pp_seq (fun f (x,e) -> fprintf f "%s=%a" (string_of_ident x) pp_can_exp e)) f eq

let pp_inst f inst =
  if !Config.verbose1 then
    if inst <> []
    then fprintf f "Instantiations inferred: %a@." pp_eq inst

(** default values *)
let can_exp_zero = Ce_num 0
let can_exp_one = Ce_num 1
let can_exp_empty_list = Ce_fun ("@list", [])
let can_exp_empty_set = Ce_fun ("@set", [])
let can_atom_false = mk_NEQ (can_exp_zero, can_exp_zero)
let can_atom_true = mk_NEQ (can_exp_zero, can_exp_one)
let can_uform_false = ([can_atom_false],[])
let can_form_false = (can_uform_false,[])

let neg_can_atom = function
  | EQ (e1,e2) -> mk_NEQ (e1,e2)
  | NEQ (e1,e2) -> mk_EQ (e1,e2)
  | POS (n,nel) -> mk_POS ((-n)-1,List.map (fun(n,e)->(-n,e)) nel)

let remove_pure_duplicates l =
  let l2 = List.stable_sort compare_pure l in
  let rec f = function
    | x::y::xs -> if compare_pure x y = 0 then f(y::xs) else x::(f(y::xs))
    | xs -> xs
  in f l2

let eval_can_atom pl ca =
  let eval_eq = function
    | (Ce_num n1, Ce_num n2) -> n1=n2
    | e1,e2 -> if equal_exp e1 e2 then true else raise Not_found in
  let eval_pos = function
    | (n,[]) -> n>=0
    | (n,nel) -> raise Not_found
  in try
      (match ca with
	 | EQ (e1,e2) -> eval_eq (e1,e2)
	 | NEQ (e1,e2) -> not (eval_eq (e1,e2))
	 | POS (n,nel) -> eval_pos (n,nel))
    with Not_found ->
      if mem_atom ca pl then true
      else
	let not_ca = neg_can_atom ca in 
	if mem_atom not_ca pl then false
	else raise Not_found

let is_EQ = function
  | EQ _ -> true
  | _ -> false

let is_NEQ = function
  | NEQ _ -> true
  | _ -> false

let is_POS = function
  | POS _ -> true
  | _ -> false

let un_EQ = function
  | EQ (e1,e2) -> (e1,e2)
  | _ -> raise (Invalid_argument "un_EQ")

let un_NEQ = function
  | NEQ (e1,e2) -> (e1,e2)
  | _ -> raise (Invalid_argument "un_NEQ")


let rec mk_EQ_exn(e1,e2) = 
  match (e1,e2) with
    | Ce_ident i, Ce_ident j -> if equal_ident i j then Some pure_true else Some (pure_one (mk_EQ (e1,e2)))
    | Ce_num n, Ce_num m     -> if n=m then Some pure_true else None
    | Ce_num _, (Ce_ident _ | Ce_fun _)
    | (Ce_ident _ | Ce_fun _), Ce_num _ -> Some (pure_one (mk_EQ (e1,e2)))
    | Ce_fun("@item",[e1]),Ce_fun("@item",[e2])-> mk_EQ_exn (e1, e2)
    | Ce_fun("@item",_),   Ce_fun("@list",el2) -> mk_list_EQ ([e1], el2)
    | Ce_fun("@item",_),   Ce_fun("@set",el2)  -> mk_set_EQ ([e1], el2)
    | Ce_fun("@list",el1), Ce_fun("@list",el2) -> mk_list_EQ (el1, el2)
    | Ce_fun("@list",el1), Ce_fun("@item",_)   -> mk_list_EQ (el1, [e2])
    | Ce_fun("@set",el1),  Ce_fun("@set",el2)  -> mk_set_EQ (el1, el2)
    | Ce_fun("@set",el1),  Ce_fun("@item",_)   -> mk_set_EQ (el1, [e2])
    | _ ->
  match mk_grp("+",0,[(1,e1);(-1,e2)]) with
    | Ce_num 0 -> Some pure_true
    | Ce_num n -> None
    | Ce_ident i -> Some (pure_one (mk_EQ (can_exp_zero, Ce_ident i)))
    | Ce_fun(s,el) -> Some (pure_one (mk_EQ (can_exp_zero, Ce_fun (s,el))))
    | Ce_grp("+",n,(1,Ce_ident i)::nel) -> 
	Some (pure_one (mk_EQ (Ce_ident i, mk_grp("+",-n,List.map (fun(n,e)->(-n,e)) nel))))
    | Ce_grp("+",n,(-1,Ce_ident i)::nel) -> 
	Some (pure_one (mk_EQ (Ce_ident i, mk_grp("+",n,nel))))
    | _ -> 
  match mk_grp("^",0,[(1,e1);(1,e2)]) with
    | Ce_num 0 -> Some pure_true
    | Ce_num n -> None
    | Ce_ident i -> Some (pure_one (mk_EQ (can_exp_zero, Ce_ident i)))
    | Ce_fun(s,el) -> Some (pure_one (mk_EQ (can_exp_zero, Ce_fun (s,el))))
    | Ce_grp("^",n,(1,e1)::nel) -> Some (pure_one (mk_EQ (e1, mk_grp("^",n,nel))))
    | e -> 
	Format.fprintf !Config.formatter "%a@." pp_can_exp e;
	assert false

and mk_list_EQ = 
  let rec go f = function [] -> Some pure_true
    | e::el -> match f e with
        None -> None
      | Some l -> go f el >>= fun l' -> Some (pure_conj l l') in
  let certainly_non_empty el = List.exists (function Ce_fun("@item",_) -> true | _ -> false)  el in
  function
  | ([], el2) -> 
	if certainly_non_empty el2 then None 
	else go (fun e -> mk_EQ_exn (e, can_exp_empty_list)) el2
  | (el1, []) ->
	if certainly_non_empty el1 then None 
	else go (fun e -> mk_EQ_exn (e, can_exp_empty_list)) el1
  | (Ce_fun("@item",[e1])::el1, Ce_fun("@item",[e2])::el2) ->
	mk_EQ_exn (e1, e2) >>= fun pl ->
	mk_list_EQ (el1, el2) >>= fun pl' ->
	Some (pure_conj pl pl')
  | ([e1], [e2]) -> mk_EQ_exn (e1,e2)
  | (e1::el1, e2::el2) when equal_exp e1 e2 -> mk_list_EQ (el1,el2)
  | (el1, el2) ->
     begin match (List.rev el1, List.rev el2) with
     | (e1::el1, e2::el2) when equal_exp e1 e2 -> mk_list_EQ (List.rev el1, List.rev el2)
     | _ -> Some (pure_one (mk_EQ (mk_fun ("@list",el1), mk_fun("@list",el2))))
     end

and mk_set_EQ (el1,el2) =
  let rec go f = function [] -> Some pure_true
    | e::el -> match f e with
        None -> None
      | Some l -> go f el >>= fun l' -> Some (pure_conj l l') in
  let certainly_non_empty el = List.exists (function Ce_fun("@item",_) -> true | _ -> false)  el in
  let (el,el1) = List.partition (fun x -> mem_exp x el2) el1 in
  let el2 = List.filter (fun x -> not (mem_exp x el)) el2 in
  match (el1,el2) with
    | ([], el2) ->
	if certainly_non_empty el2 then None 
	else go (fun e -> mk_EQ_exn (e, can_exp_empty_set)) el2
    | (el1, []) ->
	if certainly_non_empty el1 then None 
	else go (fun e -> mk_EQ_exn (e, can_exp_empty_set)) el1
    | ([e1], [e2]) -> mk_EQ_exn (e1,e2)
    | (el1, el2) -> Some (pure_one (mk_EQ (mk_fun ("@set",el1), mk_fun("@set",el2))))

let mk_NEQ_exn (e1,e2) = 
  match mk_EQ_exn (e1,e2) with
  | None -> Some pure_true
  | Some [EQ(e1,e2)] -> Some [NEQ(e1,e2)] (* Optimisation *)
  | Some res -> 
      if pure_is_true res then None else Some (pure_one (mk_NEQ(e1,e2)))

let is_uneq pl (e1,e2) =
  match mk_NEQ_exn(e1,e2) with 
  | None -> false
  | Some [ca] -> mem_atom ca pl (* Optimisation *)
  | Some res -> pure_is_true res || mem_atom (mk_NEQ (e1,e2)) pl

let normalize_can_atom ca = match ca with
  | EQ (e1,e2) -> mk_EQ_exn (e1,e2)
  | NEQ (e1,e2) -> mk_NEQ_exn (e1,e2)
  | POS(n,nel) -> 
      (match nel with
	| [] -> if n>=0 then Some pure_true else None
	| _::_ -> Some (pure_one (mk_POS(n,nel))))

let remove_exp_duplicates l =
  let l2 = List.stable_sort compare_exp l in
  let rec f = function
    | x::y::xs -> if equal_exp x y then f(y::xs) else x::(f(y::xs))
    | xs -> xs
  in f l2

let normalize_pure pl =
  let f ca res = match res with 
   | None -> None
   | Some res -> match normalize_can_atom ca with
     | None -> None
     | Some pl -> Some (pure_conj pl res)
  in
     pure_fold f pl (Some pure_true)


(** {2 Disjoint sets} *)

let dset_from_list el = 
  let el' = remove_exp_duplicates el in
  List.filter (function x -> not (equal_exp can_exp_zero x)) el'

let can_exp_is_zero e = equal_exp e can_exp_zero
let dset_emp = []
let dset_add e s = if can_exp_is_zero e then s else remove_exp_duplicates (e::s)
let dset_mem e s = (can_exp_is_zero e || List.exists (fun x -> equal_exp x e) s)
let dset_remove e s = List.filter (fun x -> not (equal_exp e x)) s
let dset_to_list s = s
let dset_inter s1 s2 = List.filter (fun x -> dset_mem x s2) s1
let dset_subset s1 s2 = List.for_all (fun x -> dset_mem x s2) s1
let dset_partition f s = List.partition f s 
let dset_notmem pl e el = List.for_all (fun e' -> is_uneq pl (e,e')) el

(* Returns the set of expresssions that are not equal to e in pl *)
let get_uneql e pl = 
  let rec go = function
    | [] -> []
    | NEQ(e1,e2)::pl when equal_exp e1 e -> e2 :: go pl
    | NEQ(e1,e2)::pl when equal_exp e2 e -> e1 :: go pl
    | _::pl -> go pl
  in go (pure_to_list pl)


(* substitutions *)
type subst = can_exp -> can_exp 

let rec mk_subst (l: (ident * can_exp) list) e =
  let rec f_e e = match e with
    | Ce_num n -> e
    | Ce_ident i ->
	let rec f = function
	  | [] -> e
	  | (i',e')::il -> if equal_ident i i' then e'
	    else f il
	in f l
    | Ce_fun (i,el) -> mk_fun (i, List.map f_e el)
    | Ce_grp(op,n,nel) -> mk_grp(op,n,List.map (fun(n,e)->(n, f_e e)) nel)
  in f_e e

let mk_single_subst (i,e) =
  mk_subst [(i,e)]

let mk_gensym_garb_subst i =
  mk_single_subst (i, Ce_ident (gensym_garb i))

let gensym_subst_list is_garb l =
  let l' = List.map (fun i -> (i,Ce_ident ((if is_garb then gensym_garb else gensym) i))) l in
  mk_subst l'

let mk_gensym_subst_list =
  gensym_subst_list false

let mk_gensym_garb_subst_list =
  gensym_subst_list true

let map_idset f_e ss =
  let f i = match f_e (Ce_ident i) with
    | Ce_ident x -> x
    | _ -> assert false
  in IdSet.fold (fun i s -> IdSet.add (f i) s) ss IdSet.empty

let map_atom f_e = function
  | EQ (e1,e2) -> mk_EQ (f_e e1, f_e e2)
  | NEQ (e1,e2) -> mk_NEQ (f_e e1, f_e e2)
  | POS (n,nel) -> mk_POS (n, List.map (fun(n,e)->(n,f_e e)) nel)

let map_pure f_e pl = pure_fold (fun ca res -> pure_conj_one (map_atom f_e ca) res) pl pure_true

let map_dset f_e ds = 
  let ds = List.map f_e ds in
  let ds = remove_exp_duplicates ds in
  List.filter (fun x -> x <> Ce_num 0) ds

let map_fld f_e fld = List.map (fun (c,e) -> (c, f_e e)) fld


let (++) = IdSet.union

let rec fv_can_exp = function
  | Ce_num _ -> IdSet.empty
  | Ce_ident i -> IdSet.singleton i
  | Ce_fun (i,el) -> List.fold_left (fun s e -> s ++ fv_can_exp e) IdSet.empty el
  | Ce_grp(op,n,nel) -> List.fold_left (fun s (_,e) -> s ++ fv_can_exp e) IdSet.empty nel 

let fv_can_atom = function
  | EQ(e1,e2) -> (fv_can_exp e1) ++ (fv_can_exp e2)
  | NEQ(e1,e2) -> (fv_can_exp e1) ++ (fv_can_exp e2)
  | POS(n,nel) -> List.fold_left (fun s (_,e) -> s ++ fv_can_exp e) IdSet.empty nel

let fv_norm_can_atom ca =
  let fv = fv_can_atom ca
  in IdSet.filter (fun x -> not (is_existential x || is_garbage x)) fv

let fv_dset ds = List.fold_left (++) IdSet.empty (List.map fv_can_exp ds)
let fv_fld fld = List.fold_left (++) IdSet.empty (List.map (function (_,ce) -> fv_can_exp ce) fld)


(** return true if of the form _x *)
let existential_id = function
  | Ce_ident id -> is_existential id
  | _ -> false

(** unify [e1] and [e2] instantiating existential variables in [e2]. *)
let unify_exp inst e1 e2 = 
  let rec go (inst,pl) e1 e2 = match (e1,e2) with
    | e1, Ce_ident i2 ->
      (try 
	let (_,e3) = List.find (fun (x,_)-> equal_ident x i2) inst in
	go (inst,pl) e1 e3
      with Not_found -> 
	if is_existential i2 then
	  let sub = mk_single_subst (i2,e1) in
	  let inst = List.map (fun (t,e) -> (t,sub e)) inst in
	  Some ((i2,e1)::inst, pl)
	else if equal_exp e1 e2 then Some (inst,pl)
	else None)
    | Ce_num n1, Ce_num n2 -> if n1 = n2 then Some (inst,pl) else None
    | Ce_fun (s1,el1), Ce_fun (s2,el2) ->
	if Pervasives.(=) s1 s2 then go_list (inst,pl) el1 el2 else None
    | e1, e2 -> if e1==e2 then Some (inst,pl) else None
  and go_list res el1 el2 = match (el1,el2) with
    | [], [] -> Some res
    | [],_::_ -> None 
    | _::_,[] -> None
    | e1::el1,e2::el2 -> 
	go res e1 e2 >>= fun res ->
	go_list res el1 el2
  in
  go (inst,pure_true) e1 e2 

let weak_unify_exp inst e1 e2 =
  let rec go inst = function None -> None | Some pl ->
    match pure_first (function 
	| EQ(Ce_ident i, _) when is_existential i -> true
	| EQ(_, Ce_ident i) when is_existential i -> true
	| _ -> false) pl with
      | None -> Some (inst, pl)
      | Some (EQ(Ce_ident i, e), pl) when is_existential i -> 
	  go ((i,e)::inst) (normalize_pure (map_pure (mk_subst [(i,e)]) pl))
      | Some (EQ(e, Ce_ident i), pl) when is_existential i ->
	  go ((i,e)::inst) (normalize_pure (map_pure (mk_subst [(i,e)]) pl))
      | _ -> assert false in
  go inst (mk_EQ_exn (e1, mk_subst inst e2))

(** {2 Fields} *)
(** Represented as a sorted association list *)

let fld_from_list cel = List.sort compare_component_exp cel
let fld_to_list cel = cel
let fld_inter cel1 cel2 = List.filter (fun ce -> mem_component_exp ce cel1) cel2 
let fld_union cel1 cel2 = List.merge compare_component_exp cel1 cel2
let fld_get t cel = List.assoc t cel
let fld_mem (t,e) cel = List.mem (t,e) cel
let fld_remove t cel = List.remove_assoc t cel
let fld_containing e0 cel = fst (List.hd (List.filter (fun (_,x) -> equal_exp x e0) cel))
let fld_exists f cel = List.exists f cel
let fld_filter f cel = List.filter f cel

let fld_set (t,e) cel = 
  let rec insert xl = match xl with
  | [] -> [(t,e)]
  | (t',e')::xl when t<t' -> (t,e)::(t',e')::xl
  | (t',e')::xl when Pervasives.(=) t t' -> (t,e)::xl
  | x::xl -> x::insert xl in
  insert cel

(** check that el1 is subset of el2 and instantiate existentials in el2 along the way *)
let fld_subset inst cel1 cel2 =
  let (cel1,cel2) = (List.sort compare_component_exp cel1, List.sort compare_component_exp cel2) in
  let rec subset inst pl = function
    | (_,[]) -> Some (inst,pl)
    | ((t1,e1)::l,(t2,e2)::l') when Pervasives.(=) t1 t2 ->
	unify_exp inst e1 e2 >>= fun (inst,pl') ->
        subset inst (pure_conj pl' pl) (l,l')
    | ((t1,e1)::l,(t2,e2)::l') when t1<t2 ->
        subset inst pl (l,(t2,e2)::l')
    | (l,(t2,e2)::l') ->
	begin match e2 with
	  | Ce_ident x when is_existential x ->
            let split_id = Ce_ident (gensym_garb (create_ident t2)) in
	    let sub = mk_single_subst (x,split_id) in
	    let l'' = List.map (fun (t,e) -> (t,sub e)) l' in
	    subset ((x,split_id)::inst) pl (l,l'')
          | _ -> None
	end
  in subset inst pure_true (cel1,cel2)

let weak_fld_subset inst cel1 cel2 =
  let (cel1,cel2) = (List.sort compare_component_exp cel1, List.sort compare_component_exp cel2) in
  let rec subset inst pl = function
    | (_,[]) -> Some (inst,pl)
    | ((t1,e1)::l,(t2,e2)::l') when Pervasives.(=) t1 t2 ->
	weak_unify_exp inst e1 e2 >>= fun (inst,pl') ->
        subset inst (pure_conj pl' pl) (l,l')
    | ((t1,e1)::l,(t2,e2)::l') when t1<t2 ->
        subset inst pl (l,(t2,e2)::l')
    | (l,(t2,e2)::l') ->
	begin match e2 with
	  | Ce_ident x when is_existential x ->
            let split_id = Ce_ident (gensym_garb (create_ident t2)) in
	    let sub = mk_single_subst (x,split_id) in
	    let l'' = List.map (fun (t,e) -> (t,sub e)) l' in
	    subset ((x,split_id)::inst) pl (l,l'')
          | _ -> None
	end
  in subset inst pure_true (cel1,cel2)



(**** equalities ****)

(* equalities are represented as a list  x1=E1,...,xn=En  where
     -the xi are distinct
     -the Ei do not contain occurrences of any xj and are normalized
     -xi < fv(Ei) where < is Pervasives.compare *)

type equalities = (ident * can_exp) list

let empty_eq = []

(* create a substitution from equalities *)
let eq_to_sub eq =
  mk_subst eq

let eq_to_can_pure eq =
  let pl = List.fold_left (fun r (id,e) -> pure_conj_one (mk_EQ(Ce_ident id,e)) r) pure_true eq in
  (*List.filter (function EQ(x,y)-> not (equal_exp x y) && not (existential_id x) && not (existential_id y) | _ -> false)*) pl

let sub_eq sub eq =
  (* fprintf !Config.formatter "Substituting into %a@." pp_eq eq; *)
  let f i =
    let un_id = function Ce_ident x -> x | _ -> assert false
    in un_id (sub (Ce_ident i)) in
  let eq' = List.map (fun (i,e) -> (f i, sub e)) eq in
  (* fprintf !Config.formatter "Returns %a@." pp_eq eq'; *)
  eq'

let mk_sub (i,e) eq = sub_eq (mk_single_subst (i,e)) eq

(* extend equalities eq with e1=e2 *)
let eq_add_unsafe eq (e1,e2) =
  let add_one ca eq = match ca with
    | EQ(Ce_ident i, e) -> ((i,e)::(mk_sub (i,e) eq))
    | EQ(e, Ce_ident i) -> ((i,e)::(mk_sub (i,e) eq))
    | _ -> raise (Failure "eq_add") in
  let (e1,e2) = 
    let sub = (eq_to_sub eq) in 
    (sub e1, sub e2) in
  mk_EQ_exn (e1,e2) >>= fun pl -> 
  Some (pure_fold add_one pl eq)

let eq_add eq (e1,e2) = try eq_add_unsafe eq (e1,e2) with Failure _ -> Some eq

(** extend equalities eq with e1=f1...en=fn *)
let eq_add_list eq pl =
  let f ca eqo = match eqo with
    | None -> None
    | Some eq ->
	match ca with
	  | EQ (e1,e2) -> eq_add eq (e1,e2)
	  | NEQ _ -> Some eq
	  | POS _ -> Some eq
  in pure_fold f pl (Some eq)

(** extend equalities eq with e1=f1...en=fn *)
let eq_partition_list eq pl =
  let f ca eqplo = match eqplo with
    | None -> None
    | Some (eq,pl) ->
	match ca with
	  | EQ (e1,e2) -> 
	    begin try eq_add_unsafe eq (e1,e2) >>= fun eq -> Some (eq,pl)
            with Failure _ -> Some (eq, pure_conj_one ca pl) end
	  | NEQ _ -> Some (eq, pure_conj_one ca pl)
	  | POS _ -> Some (eq, pure_conj_one ca pl)
  in pure_fold f pl (Some (eq,pure_true))


(* remove binding for e from eq *)
let kill_vars_eq eq kill =
  let kl = IdSet.elements kill in
  let rec f (eq_left,eq_gone) = function
    | [] -> (eq_left,eq_gone)
    | id::l ->
	let (eq1,eq2) = List.partition (fun (x,_) -> equal_ident x id) eq_left
	in match eq1 with
	  | [] -> (eq_left,eq_gone)
	  | [(_,e1)] -> f (mk_sub (id,e1) eq2, (id,e1)::(mk_sub (id,e1) eq_gone)) l
	  | _ -> assert false in
  let (eq_left,eq_gone) = f (eq,empty_eq) kl
  in (eq_left, eq_to_sub eq_gone)

let sub_equalities sub eq =
  let eq' = sub_eq sub eq in
  match List.fold_left (fun eq1o (id,e) ->
    match eq1o with
      | None -> None
      | Some eq1 ->
	  eq_add eq1 (Ce_ident id, e)) (Some []) eq' with
  | Some eq'' -> eq''
  | None -> assert false

let combine_eq eq1 eq2 =
  let (eq,eq2) = List.partition (fun x -> mem_eq x eq1) eq2 in
  let eq1 = List.filter (fun x -> not (mem_eq x eq)) eq1 in
  (eq,eq1,eq2)

(* (* orders (LEQ) *)
module ExpSet =
  Set.Make(struct
    type t = can_exp
    let compare = compare_exp
  end)

let is_leq pl (e1,e2) = equal_exp e1 e2 ||
  let bsr = ref ExpSet.empty in 
  let f n ca = match ca with
    | LEQ(Ce_num n',e') when n'<=n -> bsr := ExpSet.add e' !bsr; n
    | LEQ(e,e') when ExpSet.mem e !bsr -> 
      (match e' with 
	| Ce_num n' -> max n n'
	| _ -> bsr := ExpSet.add e' !bsr; n)
    | _ -> n
  in (* normally, I should compute the fixpoint *)
  let n = List.fold_left f min_int (LEQ(Ce_num min_int,e1)::pl) in
  match e2 with 
    | Ce_num m when n<=m -> true 
    | _ -> ExpSet.mem e2 !bsr
*)

(** Check whether [pl |- ca]. Returns [Some true] if [pl => ca];
   [Some false] if [pl => not ca]; [None] if it depends. *)
let entails_atom pl = function
  | EQ(e1,e2) -> 
      if equal_exp e1 e2 then Some true 
      else if is_uneq pl (e1,e2) then Some false
      else None
  | NEQ(e1,e2) ->
      if equal_exp e1 e2 then Some false 
      else if is_uneq pl (e1,e2) || existential_id e1 || existential_id e2 then Some true
      else None
  | POS(n,nel) -> 
      let f = function
	| POS(n',nel') -> compare_int_exp_list nel nel' = 0 && n>=n'
	| _ -> false in
      if pure_fold (fun x b -> b || f x) pl false then Some true else None (* TODO *)

exception Entails_Internal

let entails_pure_list pl pl' =
  let f ca e = match entails_atom pl ca with
    | Some true -> e
    | Some false -> raise Entails_Internal
    | None -> pure_conj_one ca e in
  let myfilter ca = match ca with
    | EQ(e1,e2) -> true
    | NEQ(e1,e2) -> not (existential_id e1) && not (existential_id e2)
    | POS(n,nel) -> true in
  try 
   Some (pure_filter myfilter (pure_fold f pl' pure_true))
  with Entails_Internal -> None

(** use equalities in [pl] and try to derive a contradiction *)
let aggr_norm_pure pl =
  eq_partition_list empty_eq pl >>= fun (eq,pl) ->
  normalize_pure (map_pure (eq_to_sub eq) pl) >>= fun pl ->
  Some (pure_conj (eq_to_can_pure eq) pl)

(** normalize list and remove equalities *)
let aggr_prove_pure pl =
  let f = function
    | EQ (e1,e2) -> 
	not (equal_exp e1 e2) && not (existential_id e1) && not (existential_id e2)
    | NEQ _ -> true
    | POS _ -> true in
  aggr_norm_pure pl >>= fun  pl ->
  Some (pure_filter f pl)


(* constraints *)
type constr = (can_exp * can_exp) list * (can_exp * can_exp)

let compare_constr (cal1,ca1) (cal2,ca2) =
  let n = compare_pure cal1 cal2
  in if n<>0 then n
    else compare_atom ca1 ca2

module ConstrSet =
  Set.Make(struct
    type t = (can_pure * can_atom)
    let compare = compare_constr
  end)

type constraints = ConstrSet.t

let pp_constraints f cs = Format.fprintf f "@[%a@]" (pp_starsep pp_const) (ConstrSet.elements cs)

let empty_constrs = ConstrSet.empty

let normalize_constr (pl,ca) =
  match normalize_pure pl with
    | None -> []
    | Some pl ->
      (*let pl = List.stable_sort compare_atom pl in*)
      match normalize_can_atom ca with
	| None -> [(pl, can_atom_false)]
	| Some pl' -> pure_fold (fun ca res -> (pl,ca)::res) pl' []

let normalize_constrs cs =
  let add x y = ConstrSet.add y x in 
  List.fold_left add ConstrSet.empty (ConstrSet.elements cs)

let add_constr cs (l,(e1,e2)) =
  let cl = normalize_constr (List.fold_left (fun res x -> pure_conj_one (mk_NEQ x) res) pure_true l, mk_NEQ (e1,e2)) in
  List.fold_left (fun x y -> ConstrSet.add y x) cs cl

let add_ineq ca pl = pure_conj_one ca pl

(* rule f(F1s,E,F2s)!=f(F1s,E',F2s) => E!=E' *)
let add_fun_ineq plr =
  let rec f d = function
    | ([],[]) -> (match d with Some ca -> ca | None -> can_atom_false) 
    | (e1::l1,e2::l2) ->
	if equal_exp e1 e2 then f d (l1,l2)
	else
	  let d =
	    (match d with
	       | Some ca -> Some can_atom_true
	       | None -> Some(mk_NEQ(e1,e2)))
	  in f d (l1,l2)
    | _ -> assert false in
  let add_neq = function
    | NEQ(Ce_fun (i1,el1), Ce_fun(i2,el2)) ->
	if Pervasives.(=) i1 i2 then plr := add_ineq (f None (el1,el2)) !plr
	else ()
    | _ -> ()
  in pure_fold (fun x _ -> add_neq x) !plr ()

let simp_constr (pl,cs) =
  let plr = ref pl in
  add_fun_ineq plr;
  let f (pl1,ca) =
    let pl1 = pure_filter (fun ca1 -> not (is_uneq pl (un_NEQ ca1))) pl1 in
      if (mem_atom ca pl1 
	  || pure_fold (fun ca b -> b || (let (e1,e2) = un_NEQ ca in equal_exp e1 e2)) pl1 false
	  || is_uneq pl (un_NEQ ca))
      then None
      else if pure_is_true pl1 then 
	(plr := add_ineq ca !plr; None)
      else Some (pl1,ca) in
  let constrl = ConstrSet.elements cs in
  let csr = ref ConstrSet.empty in
    List.iter (fun c -> csr := ConstrSet.add c !csr) (map_partial f constrl);
    (!plr,!csr)

let compare_pl_cs (pl1,cs1) (pl2,cs2) =
  let n = compare_pure pl1 pl2 in if n<>0 then n else 
  ConstrSet.compare cs1 cs2

let rec simplify_constraints x = (* constraint simplification *)
  let y = simp_constr x in
  if compare_pl_cs x y = 0 then x else simplify_constraints y

(* for debug

let pp_constraints_hook = ref (fun _ -> assert false)

let simplify_constraints x =
  let y = simplify_constraints x in
  Format.fprintf !Config.formatter "@.BEFORE SIMPLIFY@.";
  !pp_constraints_hook !Config.formatter x;
    Format.fprintf !Config.formatter "@.AFTER SIMPLIFY@.";
  !pp_constraints_hook !Config.formatter y;
  y
*)

let sub_constrs sub cs =
  let f (cal,ca) = (map_pure sub cal, map_atom sub ca) in
  let csr = ref ConstrSet.empty in
  ConstrSet.iter (fun c -> csr := ConstrSet.add (f c) !csr) cs;
  !csr

let solve_constr (eq,pl,cs) =
  let rec f (eq,pl,cs) =
    match (normalize_pure pl, normalize_constrs cs) with
      | (None,_) -> None
      | (Some pl,cs) ->
	  let (pl',cs') = simplify_constraints (pl,cs) in
	  let cal = ref pure_true in
	    ConstrSet.iter (function (pl0,ca) ->
		(match pure_to_list pl0 with 
		 | [ca0] -> let (e1,e2)=un_NEQ ca in if equal_exp e1 e2 then cal := add_ineq ca0 !cal 
		 | _ -> ()))
 		cs';
	    if not (pure_is_true !cal) then
	      begin
		let eqo =
		  let add ca eqo =
		    match eqo with
		      | None -> None
		      | Some eq -> eq_add eq (un_NEQ ca) in
		    pure_fold add !cal (Some eq) in
		  match eqo with
		    | None -> None
		    | Some eq' ->
			let sub = eq_to_sub eq' in
			let pl'' = remove_atom_duplicates (map_pure sub pl') in
			let cs'' = sub_constrs sub cs'
			in f (eq',pl'',cs'')
	      end
	    else Some (eq,pl',cs')
  in f (eq,pl,cs)

let is_inconsistent pl =
  pure_fold
    (fun p b -> b || is_NEQ p && (let (e1,e2) = un_NEQ p in equal_exp e1 e2))
    pl false

let constrset_find f cs =
  let cl = ConstrSet.elements cs
  in List.find f cl

let solve_constraints eq pl cs = (* solve constraints and generate equalities and inequalities *)
  let inconsistent_eq ca (eq,pl,cs) =
    match eq_add eq (un_NEQ ca) with
      | None -> true
      | Some eq1 ->
	  let sub = eq_to_sub eq1 in
	  let pl' = remove_atom_duplicates (map_pure sub pl) in
	  let cs' = sub_constrs sub cs in
	    match solve_constr (eq1,pl',cs') with
	      | None -> true
	      | Some (_,pl'',_) -> is_inconsistent pl'' in
    match solve_constr (eq,pl,cs) with
      | None -> None
      | Some cs' ->
	  let cur_state = ref cs' in
	  let rec loop () =
	    let (eq',pl',cs') = !cur_state in
	      if not (is_inconsistent pl') then
		try
		  let (cal,ca) = constrset_find (fun (cal,ca) -> inconsistent_eq ca !cur_state) cs'
		  in  match solve_constr (eq',add_ineq ca pl',cs') with
		    | None -> None
		    | Some cs'' -> cur_state := cs''; loop()
		with Not_found -> Some ()
	      else Some ()
	  in match loop() with
	    | None -> None
	    | Some () -> Some !cur_state

(* get expression for the component from fields. *)
let rec getcomponent c fields = 
  match fields with 
    |  (c', e)::fields -> if Pervasives.(=) c c' then Some e else getcomponent c fields
    |    [] -> None

(* list of equalities for each component that is defined in both fields1 
   and fields2 *)
let rec components_equal fields1 fields2 = 
  List.fold_left 
    (fun plo (c,e) ->
      plo >>= fun pl ->
      match getcomponent c fields2 with 
        | None -> Some pl
        | Some e1 ->
	    mk_EQ_exn (e,e1) >>= fun pl' ->
	    Some (pure_conj pl' pl))
    (Some pure_true)
    fields1

let ex_fv_can_exp e =
  let fv = fv_can_exp e in
  IdSet.filter is_existential fv  

(** Replace existentials with gensym'd normal vars. *)
let existentials_rename_tonormal_sub fv =
  let idl = IdSet.elements fv in
  let idl' = List.map (fun x -> (x, Ce_ident (gensym (un_existential x)))) idl
  in mk_subst idl'

(** Replace existentials with gensym'd existential vars. *)
let existentials_rename_sub fv =
  let idl = IdSet.elements fv in
  let idl' = List.map (fun x -> (x, Ce_ident (gensym x))) idl
  in mk_subst idl'

let fv_keep_can_atom ca = match ca with
  | EQ(ce1,ce2) ->
      let s1 = fv_can_exp ce1 in
      let s2 = fv_can_exp ce2 in
      let f s = IdSet.for_all (fun id -> not (is_existential id)) s in
      (if f s1 then s2 else IdSet.empty) ++
      (if f s2 then s1 else IdSet.empty)
  | NEQ _ -> IdSet.empty
  | POS _ -> IdSet.empty

