module StringSet =
  Set.Make(struct
    type t = string
    let compare = compare
  end)

module StringMap =
  Map.Make(struct
    type t = string
    let compare = compare
  end)

let rec iter_pairs f l1 l2 =
  let g x1 = List.iter (fun x2 -> f (x1,x2)) l2
  in List.iter g l1

let rec iter_pairwise f = function
  | [] -> ()
  | x::l -> List.iter (function y -> f (x,y)) l;
            iter_pairwise f l

let unSome = function
  | None -> assert false
  | Some x -> x

let map_partial f l =
  let rec go = function
    | [] -> []
    | x::xl -> match f x with
       | None -> go xl
       | Some y -> y :: go xl in
  go l

let map_option f l =
  let add a b =
    match a with
      | None -> None
      | Some l' ->
	  match f b with
	    | None -> None
	    | Some x -> Some (x::l')
  in match List.fold_left add (Some []) l
    with
      | None -> None
      | Some l' -> Some (List.rev l')

(* parameters for identifiers, in priority order: builtin compare used *)
type id_status = GARBAGE | RENAMED | ORIGINAL
type id_kind = EXISTENTIAL_VAR | NORMAL_VAR
type ident = {status:id_status; kind:id_kind; unique_id:int; name:string; stamp:int}

(*let compare_status s1 s2 = match s1,s2 with
  | GARBAGE,GARBAGE -> 0
  | GARBAGE,_ -> -1
  | _,GARBAGE -> 1
  | RENAMED,RENAMED -> 0
  | RENAMED,_ -> -1
  | _,RENAMED -> 1
  | ORIGINAL,ORIGINAL -> 0

let compare_kind k1 k2 = match k1,k2 with
  | EXISTENTIAL_VAR,EXISTENTIAL_VAR -> 0
  | EXISTENTIAL_VAR,_ -> -1
  | _,EXISTENTIAL_VAR -> 1
  | NORMAL_VAR,NORMAL_VAR -> 0
*)

let compare_status (s1: id_status) (s2: id_status) = Obj.magic s1 - (Obj.magic s2 : int)
let compare_kind (s1: id_kind) (s2: id_kind) = Obj.magic s1 - (Obj.magic s2 : int)

let compare_ident i1 i2 =
  if i1==i2 then 0 else
  let n = compare_status i1.status i2.status in if n<>0 then n else
  let n = compare_kind i1.kind i2.kind in if n<>0 then n else
  -(compare i1.unique_id i2.unique_id)

let abs_compare_ident i1 i2 =
  if i1==i2 then 0 else
  let n = compare_status i1.status i2.status in if n<>0 then n else
  let n = compare_kind i1.kind i2.kind in if n<>0 then n else
  if i1.kind=EXISTENTIAL_VAR then 0 else
    -(compare i1.unique_id i2.unique_id)

let equal_ident i1 i2 =
  i1==i2 || (i1.status==i2.status && i1.kind==i2.kind && i1.unique_id == i2.unique_id)

module IdSet =
  Set.Make(struct
    type t = ident
    let compare = compare_ident
  end)

module IdMap =
  Map.Make(struct
    type t = ident
    let compare = compare_ident
  end)

let gensym_count = ref StringMap.empty
let create_ident_count = ref StringMap.empty
let unique_id_count = ref 0
let init_gensym () = gensym_count := StringMap.empty

let get_unique_id () =
  let _ = incr unique_id_count
  in !unique_id_count

(* generate fresh identifiers *)
let gensym id =
  let count = try StringMap.find id.name !gensym_count
  with Not_found -> -1
  in
    gensym_count := StringMap.add id.name (count-1) !gensym_count;
    {id with status=RENAMED; stamp = count; unique_id=get_unique_id()}

let gensym_garb id =
  {(gensym id) with kind=EXISTENTIAL_VAR; status = GARBAGE}

let tempid () = gensym {status= RENAMED; kind= NORMAL_VAR; name= "temporary"; stamp=0; unique_id=0}

let existential_name s = String.get s 0 = '_'

let un_EX s =
    try String.sub s 1 ((String.length s)- 1)
    with Not_found -> assert false

let is_existential id =
  id.kind = EXISTENTIAL_VAR

let is_garbage id =
  id.status = GARBAGE

let un_existential id =
  assert (id.kind = EXISTENTIAL_VAR);
  {id with kind = NORMAL_VAR}

let into_existential id =
  {id with kind = EXISTENTIAL_VAR}

let create_ident s =
  let (name,kind) = if existential_name s
  then (un_EX s, EXISTENTIAL_VAR)
    else (s,NORMAL_VAR) in
  let unique_id = try StringMap.find s !create_ident_count
    with Not_found ->
      let n = get_unique_id() in
      let _ =  create_ident_count := StringMap.add s n !create_ident_count
      in n
  in {status=ORIGINAL; stamp=0; name=name; kind=kind; unique_id=unique_id}


let string_of_ident id =
  (if id.kind = EXISTENTIAL_VAR then "_" else "") ^ 
    (if id.status=GARBAGE then "%" (* "%" *) else "") ^
    id.name ^
    (if id.stamp = 0 then "" else "_" ^ string_of_int (-1 - id.stamp))

let pp_ident f id = Format.fprintf f "%s" (string_of_ident id)

let rec pp_listsep pp s1 s2 f = function
  | [] -> Format.fprintf f "%s" s2
  | [x] -> Format.fprintf f "%a" pp x
  | x::l -> Format.fprintf f "%a%s %a" pp x s1 (pp_listsep pp s1 s2) l

let pp_idset f ids = Format.fprintf f "{%a}" (pp_listsep pp_ident "," "") (IdSet.elements ids)

let tid_var = create_ident "TID"
let mytid_var = create_ident "myTID" (* should appear only in actions *)
