open Misc
open Parsetree
open Exp
open Symbtree
open Inductive_preds
open Convert

let (>>) f g x = g (f x)


(** {2 Error handling} *)

exception Vcgen_error of string * Location.t

let err_num = ref 0

let error_handler do_f =
  let res = 
    try 
      do_f ()
    with 
       | Conversion_error (s, loc) ->
	  print_endline ("\n" ^ Location.sprint loc ^ "Syntax error: " ^ s);
	  invalid_arg ("Errors found in input.\n")
       | Vcgen_error (s, loc) ->
	  print_endline ("\n" ^ Location.sprint loc ^ s);
	  invalid_arg ("Errors found in input.\n")
  in
  if err_num.contents <> 0 then invalid_arg ("Errors found in input.\n");
  res

let error (s,loc) = raise (Vcgen_error (s, loc))

let warning (s,loc) = print_endline ("\n" ^ Location.sprint loc ^ "Warning: " ^ s)


let argnum_error (s1,s2,l1,l2,loc) = 
  error (s1 ^ " has " ^ string_of_int (List.length l2) ^ s2 ^ " arguments, but "
	 ^ string_of_int (List.length l1) ^ " are expected.", loc)

(** [empty_or_error x (s1,s2,loc)] checks that the set [x] is empty. *) 
let empty_or_error x (s1,s2,loc) = 
  if not (IdSet.is_empty x) then error (s1 ^ string_of_ident (IdSet.choose x) ^ s2, loc)


(** {2 Set operations} *)

let (++) = IdSet.union
let (--) = IdSet.diff
let (+) s x = s ++ (IdSet.singleton x)
let (-) s x = IdSet.remove x s
let set_of_list l =
  List.fold_left (+) IdSet.empty l
let union_map f l =
   List.fold_left (fun ss x -> ss ++ (f x)) IdSet.empty l

(** {2 Environment} *)

type typ = 
  | Pty_null
  | Pty_base of string
  | Pty_class of string 

let string_of_ty = function
  | Pty_null -> "null_t"
  | Pty_base s -> s
  | Pty_class s -> s

let ty_of_string s = 
  if List.mem s ["void"; "int"; "bool"; "thread_id"; "set"; "list"] then Pty_base s else Pty_class s

(* User-defined structures *)
type class_item =
    { cla_fields: (typ * Location.t) StringMap.t;
      cla_loc: Location.t }

type function_item =
    { fn_resty: typ;
      param: (ident * typ * Location.t) list * (ident * typ * Location.t) list;
      locals: (ident * typ * Location.t) list;
      bound : IdSet.t;  (* locals + params *)
      mutable calls: StringSet.t StringMap.t;  (* map from fid to minimum set of resources 
                                                  always acquired before calling fid *)
      mutable in_parallel_with: StringSet.t;
      mutable requires: StringSet.t;
      mutable writes_heap: bool;
      mutable modif: IdSet.t;
      mutable vars: IdSet.t;
      pre: can_prop;
      body: p_statement list;
      post: can_prop;
      fn_loc: Location.t }

type action_item =
    { act_id: string;
      act_param: (string * Location.t) list;
      act_locals: IdSet.t;
      act_bound: IdSet.t; (* locals + params *)
      act_ctx: can_prop;
      act_pre: can_prop;
      act_post: can_prop;
      act_loc: Location.t }

type resource_item =
    { owned: IdSet.t;
      rvars: IdSet.t; (* Free variables of the resource invariant *)
      res_locals : (ident * typ * Location.t) list;
      res_body : p_statement list;
      rinv : can_prop option;  (* Resource invariant *)
      ractions: action_item StringMap.t; (* Resource actions *)
      res_loc: Location.t }

type env = {classes: (string,class_item) Hashtbl.t;
            mutable variables: (typ * Location.t) IdMap.t;
            functions: (string,function_item) Hashtbl.t;
	    mutable env_funcs: string list; (* in reverse order wrt the original program *)
            resources: (string,resource_item) Hashtbl.t;
	    mutable env_rsrcs: string list; (* in reverse order wrt the original program *)
	    mutable enable :  IdSet.t * IdSet.t -> StringSet.t}

let initial_env () =
  let classes = Hashtbl.create 16 in
  Hashtbl.add classes "any" {cla_fields=StringMap.empty; cla_loc=Location.none};
  { classes = classes;
    variables = IdMap.add mytid_var (Pty_base "thread_id",Location.none) 
                (IdMap.add tid_var (Pty_base "thread_id",Location.none) IdMap.empty);
    functions = Hashtbl.create 16; env_funcs = [];
    resources = Hashtbl.create 4;  env_rsrcs = [];
    enable = fun _ -> assert false}

exception Env_not_found

let lookup_class env x = try Hashtbl.find env.classes x with Not_found -> raise Env_not_found

let lookup_function env x = try Hashtbl.find env.functions x with Not_found -> raise Env_not_found

let lookup_resource env x = try Hashtbl.find env.resources x with Not_found -> raise Env_not_found

let lookup_action res x = try StringMap.find x res.ractions with Not_found -> raise Env_not_found

let env_add_fun env x fi =
  if Hashtbl.mem env.functions x then 
    error ("Function " ^ x ^ " is already defined.", fi.fn_loc)
  else begin
    Hashtbl.add env.functions x fi;
    env.env_funcs <- x :: env.env_funcs
  end

let env_add_res env x ri =
  if Hashtbl.mem env.resources x then 
    error ("Resource " ^ x ^ " is already defined.", ri.res_loc)
  else begin
    let ri_vars = ri.owned ++ ri.rvars in
    empty_or_error (IdSet.filter (fun x -> not (IdMap.mem x env.variables)) ri_vars)
	("Undeclared variable ", " in resource declaration.", ri.res_loc);
    Hashtbl.iter
      (fun y r ->
        empty_or_error (IdSet.inter r.owned ri_vars) 
	  ("Variable ", " is already owned by resource " ^ y ^ ".", ri.res_loc);
        empty_or_error (IdSet.inter r.rvars ri.owned)
	  ("Variable ", " is already used by resource " ^ y ^ ".", ri.res_loc)
      ) env.resources;
    Hashtbl.add env.resources x ri;
    env.env_rsrcs <- x :: env.env_rsrcs
  end


(** {2  Type checking} *)

let ty_unify = function
 | (ty1,ty2) when ty1=ty2 -> Some ty1
 | (Pty_null, ((Pty_base "int" | Pty_base "thread_id" | Pty_class _) as ty)) -> Some ty
 | (((Pty_base "int" | Pty_base "thread_id" | Pty_class _) as ty), Pty_null) -> Some ty
 | (Pty_class "any", (Pty_class _ as ty)) -> Some ty
 | ((Pty_class _ as ty), Pty_class "any") -> Some ty
 | (Pty_base "item", (Pty_base ("set"|"list") as ty)) -> Some ty
 | ((Pty_base ("set"|"list") as ty), Pty_base "item") -> Some ty
 | (ty1,ty2) -> None

let expect_type loc ty1 ty2 =
  match ty_unify (ty1,ty2) with
  | Some _ -> ()
  | None -> error ("Type error: expecting type `" ^ string_of_ty ty1 ^ "', but type `" ^ string_of_ty ty2 ^ "' found.", loc)

let type_of_var loc env x = 
  try fst (IdMap.find (create_ident x) env.variables)
  with Not_found -> 
    error ("Variable " ^ x ^ " has not been declared, or is out of scope.", loc)

let lookup_class_from_type loc env ty = match ty with
  | Pty_class t -> 
     begin 
       try lookup_class env t
       with Env_not_found -> error("Reference to undeclared class " ^ t ^ ".", loc)
     end
  | _ -> error ("Type error: expecting a pointer type.", loc)

let rec type_of_pexp env e = match e.pexp_desc with
  | Pexp_ident x ->  type_of_var e.pexp_loc env x
  | Pexp_num i -> if i=0 then Pty_null else Pty_base "int"
  | Pexp_bool _ -> Pty_base "bool"
  | Pexp_prefix (s,e1) -> 
      begin match s with
	| ("+" | "-") ->
	    check_pexp env (Pty_base "int") e1; Pty_base "int"
	| ("!") ->
	    check_pexp env (Pty_base "bool") e1; Pty_base "bool"
	| _ -> error("Error: found unknown unary operator (" ^ s ^ ").", e.pexp_loc)
     end
  | Pexp_infix (s,e1,e2) ->
      begin match s with
	| ("+" | "-" | "*" | "/" | "%" | "^") ->
	    check_pexp env (Pty_base "int") e1;
	    check_pexp env (Pty_base "int") e2;
	    Pty_base "int"
	| ("==" | "!=") ->
	    let ty1 = type_of_pexp env e1 in
	    let ty2 = type_of_pexp env e2 in
	    begin match ty_unify (ty1,ty2) with 
	      | None -> error ("Type error: Incompatible types in equality.", e.pexp_loc)
	      | Some _ -> ()
	    end;
	    Pty_base "bool"
	| ("<" | "<=" | ">" | ">=") ->
	    check_pexp env (Pty_base "int") e1;
	    check_pexp env (Pty_base "int") e2;
	    Pty_base "bool"
	| ("&&" | "||") ->
	    check_pexp env (Pty_base "bool") e1;
	    check_pexp env (Pty_base "bool") e2;
	    Pty_base "bool"
	| _ -> error("Error: found unknown binary operator (" ^ s ^ ").", e.pexp_loc)
      end
  | Pexp_cast (e1,ty) ->
      let _ = type_of_pexp env e1 in
      ty_of_string ty
  | Pexp_fld (e1,s) ->
      let ty = type_of_pexp env e1 in
      let ci = lookup_class_from_type e.pexp_loc env ty in
      begin try fst (StringMap.find s ci.cla_fields)
      with Not_found ->
	error("Type error: expression has type " ^ string_of_ty ty 
		^ ", which does not have field " ^ s ^ ".", e1.pexp_loc)
      end
  | Pexp_new s ->
      let ty = Pty_class s in
      ignore (lookup_class_from_type e.pexp_loc env ty); ty
  | Pexp_fun(s,el) ->
      begin match s with
	| "@item" -> 
	    if List.length el <> 1 then argnum_error("Function call","",[()],el,e.pexp_loc);
	    List.iter (type_of_pexp env >> ignore) el; 
	    Pty_base "item"
	| "@list" -> List.iter (check_pexp env (Pty_base "list")) el; Pty_base "list" 
	| "@set"  -> List.iter (check_pexp env (Pty_base "set")) el; Pty_base "set" 
	| "@zero" -> List.iter (type_of_pexp env >> ignore) el; Pty_null
	| _ -> error("Error: found unknown pure function (" ^ s ^ ").", e.pexp_loc)
      end
  | Pexp_fcall (fid,(vl,el)) ->
      let fi = 
	try lookup_function env fid
        with Env_not_found -> error ("Error: Function " ^ fid ^ " has not been defined.", e.pexp_loc) in
      check_pexp_list ("Function call", " reference", e.pexp_loc) env (fst fi.param) (List.map (fun (v,l) -> {pexp_desc=Pexp_ident v; pexp_loc=l}) vl);
      check_pexp_list ("Function call", " value", e.pexp_loc) env (snd fi.param) el;
      fi.fn_resty
 
and check_pexp env ty e =
  let ty1 = type_of_pexp env e in
  expect_type e.pexp_loc ty ty1

and check_pexp_list (s1,s2,loc) env tyl el =
  if List.length tyl = List.length el then
    List.iter2 (fun (_,ty,_) e -> check_pexp env ty e) tyl el
  else
    argnum_error(s1,s2,tyl,el,loc)

(** Check an action annotation. *)
let check_action env rid actid argl loc =
  let ri =
    try lookup_resource env rid with Env_not_found -> 
    error ("Error: The resource "^rid^" has not been declared.", loc) in
  if actid <> "" then begin
    let ai = 
      try lookup_action ri actid with Env_not_found ->
      error ("Error: The definition of resource "^rid^" does not have the action "^actid^".", loc) in
    if List.length ai.act_param <> List.length argl then
      argnum_error ("Action annotation","",ai.act_param,argl,loc)
  end

(* Make function call code *)
let mk_fcall_stm loc (reso,f,args) = 
  let e = {pexp_desc = Pexp_fcall(f,args); pexp_loc = loc} in
  match reso with
    | None -> {pstm_desc = Pstm_exp(e); pstm_loc = loc}
    | Some x -> {pstm_desc= Pstm_assign(x,e); pstm_loc = loc}

(** Check variables in the proposition [p] are defined. *)
let tycheck_prop env loc s p = 
  let g v =
    if not (IdMap.mem v env.variables) then
      error ("Error: The " ^ s ^ " contains variable " ^ string_of_ident v
      	^ " is undeclared or out of scope.", loc) in
  IdSet.iter g (fv_norm_can_prop p)

(** Simple type checking *)
let tycheck_cmd env fn_resty c =
  let rec go_stmt in_loop s = match s.pstm_desc with
    | Pstm_exp e -> 
	begin match type_of_pexp env e with
	| Pty_base "void" -> ()
	| ty -> warning("The program ignores value of type `" ^ string_of_ty ty ^ "'.", s.pstm_loc)
	end
    | Pstm_assign(v,e) -> check_pexp env (type_of_var s.pstm_loc env v) e
    | Pstm_fldassign(e1,c,e2) -> check_pexp env (type_of_pexp env {pexp_desc=Pexp_fld(e1,c); pexp_loc=s.pstm_loc}) e2
    | Pstm_dispose e -> ignore (lookup_class_from_type e.pexp_loc env (type_of_pexp env e))
    | Pstm_block sl -> List.iter (go_stmt in_loop) sl
    | Pstm_if (eo,s1,s2) ->
	(match eo with None -> () | Some e -> check_pexp env (Pty_base "bool") e); 
        go_stmt in_loop s1;
	go_stmt in_loop s2
    | Pstm_while (po,eo,s1) ->
	(match po with None -> () | Some p -> tycheck_prop env s.pstm_loc "loop invariant" (prop_to_can_prop p s.pstm_loc));
	(match eo with None -> () | Some e -> check_pexp env (Pty_base "bool") e);
        go_stmt true s1
    | Pstm_withres (rid,e,s,actid,el) ->
	check_pexp env (Pty_base "bool") e;
	check_action env rid actid el s.pstm_loc;
	List.iter (type_of_pexp env >> ignore) el;
	go_stmt in_loop s
    | Pstm_action (s,rid,actid,el) ->
	check_action env rid actid el s.pstm_loc;
	List.iter (type_of_pexp env >> ignore) el;
	go_stmt in_loop s
    | Pstm_parblock par_calls ->
	List.iter (mk_fcall_stm s.pstm_loc >> go_stmt false) par_calls
    | Pstm_return eo ->
	begin match eo with 
	| None -> 
	    if fn_resty <> Pty_base "void" then 
	      error ("Error: The return type of this function is " ^ string_of_ty fn_resty, s.pstm_loc)
	| Some e -> check_pexp env fn_resty e 
	end
    | Pstm_break ->     if not in_loop then error ("Syntax error: `break' must be inside a loop.", s.pstm_loc) 
    | Pstm_continue ->  if not in_loop then error ("Syntax error: `continue' must be inside a loop.", s.pstm_loc) 
  in
  List.iter (go_stmt false) c


let tycheck_res env ri =
  let gvars = env.variables in
  let add_var r (v,ty,loc) =
    if IdMap.mem v r then error ("Variable " ^ string_of_ident v ^ " has already been defined and is in scope.", loc)
    else IdMap.add v (ty,loc) r in
  let ff x res = List.fold_left add_var res x in
  (* add locals *)
  env.variables <- ff ri.res_locals env.variables; 
  tycheck_cmd env (Pty_base "void") ri.res_body;
  (* remove locals + params *)
  env.variables <- gvars

let tycheck_fun env fi =
  let gvars = env.variables in
  let add_var r (v,ty,loc) =
    if IdMap.mem v r then error ("Variable " ^ string_of_ident v ^ " has already been defined and is in scope.", loc)
    else IdMap.add v (ty,loc) r in
  let ff x res = List.fold_left add_var res x in
  (* add params *)
  env.variables <- ff (snd fi.param) (ff (fst fi.param) gvars);
  tycheck_prop env fi.fn_loc "precondition" fi.pre;
  (* add Result *) 
  if fi.fn_resty <> Pty_base "void" then 
    env.variables <- add_var env.variables (create_ident "Result",fi.fn_resty,Location.none);
  tycheck_prop env fi.fn_loc "postcondition" fi.post;
  (* add locals *)
  env.variables <- ff fi.locals env.variables; 
  tycheck_cmd env fi.fn_resty fi.body;
  (* remove locals + params *)
  env.variables <- gvars

(* translate an invariant to a proposition *)
let inv_to_prop po loc = match po with
  | None -> can_prop_empty
  | Some p -> prop_to_can_prop p loc

type mod_var_req_modh =
  { md: IdSet.t; (* Variables modifies *)
    vr: IdSet.t; (* Variables accessed *)
    rq: StringSet.t; (* Requires resources *)
    mh: bool         (* Modifies heap? *) }

let mvr_empty =
  { md = IdSet.empty; vr = IdSet.empty;
    rq = StringSet.empty; mh = false }

let (++.) mvr1 mvr2 =
  { md = mvr1.md ++ mvr2.md;
    vr = mvr1.vr ++ mvr2.vr;
    rq = StringSet.union mvr1.rq mvr2.rq;
    mh = mvr1.mh || mvr2.mh }

let mvr_union_map f l =
   List.fold_left (fun mvr x -> mvr ++. f x) mvr_empty l

let mvr_exp env e = 
  let rec go e = match e.pexp_desc with
    | Pexp_new _
    | Pexp_bool _
    | Pexp_num _ -> mvr_empty
    | Pexp_ident x ->
	let x = create_ident x in
	if is_existential x || is_garbage x then mvr_empty
	else {mvr_empty with vr = IdSet.singleton x}
    | Pexp_cast (e1,_)
    | Pexp_prefix (_,e1)
    | Pexp_fld (e1,_) -> go e1
    | Pexp_infix (_,e1,e2) -> go e1 ++. go e2
    | Pexp_fun (i,el) -> mvr_union_map go el
    | Pexp_fcall (fid,(vl,el)) ->
	let fi = lookup_function env fid in
	let md = List.fold_left (fun vars (v,_) -> IdSet.add (create_ident v) vars) IdSet.empty vl in
	{ mvr_empty with md = md; vr = md }
	++. mvr_union_map go el
	++. {md = fi.modif; vr = fi.vars; rq = fi.requires; mh = fi.writes_heap} in
  let mvr = go e in
  { mvr with rq = env.enable(mvr.md, mvr.vr) }

(* modified and vars and required for a sequence of statements *)
let mvr_body env stl =
  let f_e e = mvr_exp env e in
  let f_oe e = match e with None -> mvr_empty | Some e -> mvr_exp env e in
  let f_p p = {mvr_empty with vr = fv_norm_can_prop (inv_to_prop p Location.none)} in
  let rec f st = match st.pstm_desc with
    | Pstm_exp e -> mvr_exp env e
    | Pstm_assign (i,e) ->
	let id = IdSet.singleton (create_ident i) in
	mvr_exp env e ++. {mvr_empty with md=id; vr=id}
    | Pstm_fldassign (e1,_,e2) -> {(mvr_exp env e1 ++. mvr_exp env e2) with mh=true}
    | Pstm_dispose e -> {(mvr_exp env e) with mh=true}
    | Pstm_block stl -> mvr_union_map f stl
    | Pstm_if (a,st1,st2) -> f_oe a ++. f st1 ++. f st2
    | Pstm_while (inv,a,st) -> f_oe a ++. f_p inv ++. f st
    | Pstm_withres (rid,a,st,aid,argl) ->
        let ri = lookup_resource env rid in
        let mvr = f st ++. f_e a in
	(* calculate rq properly *)
	let mvr = { mvr with rq = env.enable(mvr.md, mvr.vr) } in
	let md = mvr.md -- ri.owned in
        {md = md; vr = md ++ (mvr.vr -- ri.rvars); 
	 rq = StringSet.remove rid mvr.rq; mh=mvr.mh}
    | Pstm_action (st,rid,aid,argl) ->
        f st
    | Pstm_parblock par_calls ->
	mvr_union_map f (List.map (mk_fcall_stm st.pstm_loc) par_calls)
    | Pstm_return eo -> f_oe eo
    | (Pstm_break | Pstm_continue) -> mvr_empty
  in 
  let mvr = mvr_union_map f stl in
  { mvr with rq = env.enable(mvr.md, mvr.vr) }

let (+++) sm (i,s) =
  try let s' = StringMap.find i sm
      in StringMap.add i (StringSet.inter s s') sm
  with Not_found -> StringMap.add i s sm

let (++++) sm1 sm2 =
  StringMap.fold (fun i acq sm -> sm +++ (i,acq)) sm1 sm2

(* initial call graph: only functions called directly *)
let initial_call_graph env _ fi  =
  let check_fcall fid (vl,el) loc =
    let fi = 
      try lookup_function env fid
      with Env_not_found -> error ("Function " ^ fid ^ " has not been declared.", loc) in
    if List.length vl <> List.length (fst fi.param) then
      argnum_error ("Function call", " reference", fst fi.param, vl, loc);
    if List.length el <> List.length (snd fi.param) then
      argnum_error ("Function call", " value", snd fi.param, el, loc) in
  let check_withres rid loc =
    try ignore (lookup_resource env rid)
    with Env_not_found -> error ("Resource " ^ rid ^ " has not been defined.", loc) in
  let add_fcall fid acquired = fi.calls <- fi.calls +++ (fid,acquired) in
  let rec f_e acquired e = match e.pexp_desc with
    | Pexp_new _
    | Pexp_num _ 
    | Pexp_bool _ 
    | Pexp_ident _ -> ()
    | Pexp_cast (e1,_)
    | Pexp_prefix (_,e1)
    | Pexp_fld (e1,_) -> f_e acquired e1
    | Pexp_infix (_,e1,e2) -> f_e acquired e1; f_e acquired e2
    | Pexp_fun (_,el) ->  List.iter (f_e acquired) el
    | Pexp_fcall (i,(vl,el)) ->
	check_fcall i (vl,el) e.pexp_loc;
	add_fcall i acquired;
        begin match List.filter (fun (x,_) -> IdMap.mem (create_ident x) env.variables) vl with
	  | [] -> ()
	  | (x,loc)::_ -> error ("Global variable " ^ x ^ " cannot be passed by reference.", loc)
	end;
	List.iter (f_e acquired) el in
  let rec check_pure e = match e.pexp_desc with
    | Pexp_new _
    | Pexp_fld _ 
    | Pexp_fcall _ -> error ("Arguments of a parallel call must be pure.",e.pexp_loc)
    | Pexp_num _ 
    | Pexp_bool _ 
    | Pexp_ident _ -> ()
    | Pexp_cast (e1,_)
    | Pexp_prefix (_,e1) -> check_pure e1;
    | Pexp_infix (_,e1,e2) -> check_pure e1; check_pure e2
    | Pexp_fun (_,el) ->  List.iter check_pure el in
  let rec f acquired st = match st.pstm_desc with
    | Pstm_if (_,st1,st2) -> f acquired st1; f acquired st2
    | Pstm_while (_,_,st) -> f acquired st
    | Pstm_block stl -> List.iter (f acquired) stl
    | Pstm_withres (rid,_,st,_,_) ->
        check_withres rid st.pstm_loc;
	f (StringSet.add rid acquired) st
    | Pstm_action (st,_,_,_) -> f acquired st
    | Pstm_parblock par_calls ->
	List.iter (fun (reso,i,args) ->
	  check_fcall i args st.pstm_loc;
	  add_fcall i acquired;
	  List.iter check_pure (snd args)) par_calls
    | Pstm_break
    | Pstm_continue
    | Pstm_return None -> ()
    | Pstm_exp e
    | Pstm_return (Some e)
    | Pstm_assign (_,e) 
    | Pstm_dispose e -> f_e acquired e 
    | Pstm_fldassign (e1,_,e2) -> f_e acquired e1; f_e acquired e2
  in List.iter (f StringSet.empty) fi.body

(* compute the entire call graph of fi assuming _.calls in stop are complete *)
let call_graph_fid env fi stop =
  let stopr = ref stop in
  let rec f acquired fi =
    StringMap.fold
    (fun fid' acquired' sm ->
       let fi' = lookup_function env fid' in
       let acq = StringSet.union acquired acquired' in
       let deep_calls =
         if StringSet.mem fid' !stopr
         then StringMap.map (fun acquired'' -> StringSet.union acquired'' acq) fi'.calls
         else
           (stopr := StringSet.add fid' !stopr;
           f (StringSet.union acquired acquired') fi')
       in sm ++++ (deep_calls +++ (fid',acq)))
    fi.calls StringMap.empty
  in f StringSet.empty fi

(* entire call graph of all the functions *)
let calc_call_graph env =
  let fil = env.functions in
  let _ = Hashtbl.iter (initial_call_graph env) fil in
  let stopr = ref StringSet.empty in
  let f x fi =
    fi.calls <- call_graph_fid env fi !stopr;
    stopr := StringSet.add x !stopr
  in Hashtbl.iter f fil

(* create a function  enable (mS aS) returning the resources required to modify mS and access aS *)
let mk_enable_fun env =
  let m = ref IdMap.empty in
  let find id =
    try IdMap.find id !m
    with Not_found -> (StringSet.empty,StringSet.empty) in
  let add_mods rid id =
    let (mods,vars) = find id in
     m := IdMap.add id (StringSet.add rid mods,vars) !m  in
  let add_vars rid id =
    let (mods,vars) = find id in
    m := IdMap.add id (mods,StringSet.add rid vars) !m in
  let f x ri =
    IdSet.iter (add_mods x) ri.rvars;
    IdSet.iter (add_vars x) ri.owned in
  Hashtbl.iter f env.resources;
  let enable (mS,aS) =
    StringSet.union
      (IdSet.fold (fun id s -> StringSet.union s (fst (find id))) mS StringSet.empty)
      (IdSet.fold (fun id s -> StringSet.union s (snd (find id))) aS StringSet.empty)
  in enable


(* calculates the modified and vars and required set for each function *)
let calc_mvr env =
  let enable = mk_enable_fun env in
  env.enable <- enable; 
  let shallow_mod _ fi =
    let mvr = mvr_body env fi.body in
    let fv_spec = (fv_norm_can_prop fi.pre ++ fv_norm_can_prop fi.post) in
    fi.modif <- mvr.md -- fi.bound;
    fi.vars <- (mvr.vr ++ fv_spec) -- fi.bound;
    fi.requires <- StringSet.union mvr.rq (enable (IdSet.empty, fv_spec -- fi.bound));
    fi.writes_heap <- mvr.mh;
  in Hashtbl.iter shallow_mod env.functions;
  let deep_mod _ fi =
    let call_set fid' fi' =
      StringMap.fold (fun fid _ s -> StringSet.add fid s) fi'.calls (StringSet.singleton fid') in
    let in_par_with loc = function
      | Pstm_parblock par_calls ->
          let add_call_set_to_function cs f = 
            let fi = lookup_function env f in 
            fi.in_parallel_with <- StringSet.union fi.in_parallel_with cs in
	  let rec go cs1 csl2 = match csl2 with 
	    | [] -> ()
	    | cs::csl3 -> 
                StringSet.iter (add_call_set_to_function cs) cs1;
                List.iter (StringSet.iter (add_call_set_to_function cs)) csl3;
		go (StringSet.union cs1 cs) csl3 in
	  go StringSet.empty (List.map (fun (_,f,_) -> call_set f (lookup_function env f)) par_calls)
      | _ -> ()
    in Parsetree.iter in_par_with fi.body;
    let owned_set acq =
      StringSet.fold (fun rid s -> (lookup_resource env rid).owned ++ s) acq IdSet.empty in
    let modif =
      StringMap.fold
      (fun fid acq modif -> modif ++ ((lookup_function env fid).modif -- owned_set acq))
      fi.calls fi.modif
    and vars =
      StringMap.fold
      (fun fid acq vars -> vars ++ ((lookup_function env fid).vars -- owned_set acq))
      fi.calls fi.vars
    and requires =
      StringMap.fold
      (fun fid acq req -> StringSet.union req (StringSet.diff (lookup_function env fid).requires acq))
      fi.calls fi.requires
    and writes_heap =
      StringMap.fold
      (fun fid acq writes_heap -> writes_heap || (lookup_function env fid).writes_heap)
      fi.calls fi.writes_heap
    in fi.requires <- requires;
       fi.modif <- modif;
       fi.vars <- vars;
       fi.writes_heap <- writes_heap in
    Hashtbl.iter deep_mod env.functions

(** Convert a p_action to an action_item *)
let process_action env acts (s,params,r,p,q,loc) = 
  if StringMap.mem s acts then error ("Action " ^ s ^ " is already defined.", loc);
  let r = prop_to_can_prop r loc in
  let p = prop_to_can_prop p loc in
  let q = prop_to_can_prop q loc in
  let fv = fv_norm_can_prop r ++ fv_norm_can_prop p ++ fv_norm_can_prop q in
  let exfv = ex_fv_can_prop r ++ ex_fv_can_prop p ++ ex_fv_can_prop q in
  let bound = set_of_list (List.map (fst >> create_ident) params) ++ exfv in
  empty_or_error (IdSet.filter (fun x -> not (IdMap.mem x env.variables)) (fv -- bound))
    ("Variable ", " must either be global or bound by the action.", loc);
  StringMap.add s
    {act_id=s; act_param=params; act_locals=exfv; act_bound=bound; act_ctx=r; act_pre=p; act_post=q; act_loc=loc}
    acts

let rec process_declarations iteml env =
  (* First pass -- class definitions and global variables *)
  let go1 = function
    | Pdec_class (sid,fields,loc) ->
	let fld = List.fold_left (fun res (x,ty,loc) ->
	  if StringMap.mem x res then error ("Error: Duplicate definition of field " ^ x ^ " in class " ^ sid ^ ".", loc);
	  StringMap.add x (ty_of_string ty,loc) res) StringMap.empty fields in
	Hashtbl.add env.classes sid {cla_fields = fld; cla_loc = loc}
    | Pdec_var (x,ty,loc) ->
	let id = create_ident x in
	if IdMap.mem id env.variables then 
	  error ("Global variable " ^ x ^ " is already defined.", loc);
	env.variables <- IdMap.add id (ty_of_string ty,loc) env.variables
    | Pdec_fun _ | Pdec_resource _ | Pdec_indpred _ -> () in
  (* Second pass -- predicates, functions, resources *)
  let go2 = function
    | (Pdec_class _ | Pdec_var _) -> ()
    | Pdec_fun (fid,resty,param,(pre,post),(locals,body),loc) ->
        let pre = inv_to_prop pre loc in
	let post = inv_to_prop post loc in
	let f = List.map (fun (x,y,loc) -> (create_ident x,ty_of_string y,loc)) in
	let locals = f locals in
	let param = (f (fst param), f (snd param)) in
        let bound = 
          (List.map (fun (x,_,_) -> x) >> set_of_list) (locals @ fst param @ snd param) in
	env_add_fun env fid
	    {fn_resty=ty_of_string resty; param=param; locals=locals; bound=bound;
	     calls=StringMap.empty; in_parallel_with=StringSet.empty; writes_heap=false;
	     requires=StringSet.empty; modif=IdSet.empty; vars=IdSet.empty;
	     pre=pre; body=body; post=post; fn_loc=loc}
    | Pdec_resource (rid,owned,inv,(locals,body),actions,loc) ->
	let rec modvars r st = match st.pstm_desc with
	  | (Pstm_exp _ | Pstm_fldassign _ | Pstm_dispose _
	     | Pstm_break | Pstm_continue | Pstm_return _) -> r
	  | Pstm_assign (i,_) -> IdSet.add (create_ident i) r
	  | Pstm_block stl -> List.fold_left modvars r stl
	  | Pstm_if (_,st1,st2) -> modvars (modvars r st1) st2
	  | Pstm_while (_,_,st) -> modvars r st
	  | Pstm_withres _ | Pstm_action _ | Pstm_parblock _ -> r (* TODO *) in 
        let owned = set_of_list (List.map (fst >> create_ident) owned) in
        let (rinv,rvars) = match inv with
	  | None ->
	      let locals = set_of_list (List.map (fun (x,_,_) -> create_ident x) locals) in
	      (None, IdSet.diff (List.fold_left modvars IdSet.empty body) locals)
	  | Some inv ->
	      let inv = prop_to_can_prop inv loc in
	      (Some inv, fv_norm_can_prop inv) in
        let acts = List.fold_left (process_action env) StringMap.empty actions in 
	let locals = List.map (fun (x,y,loc) -> (create_ident x,ty_of_string y,loc)) locals in
	env_add_res env rid
          {owned=owned; rvars=rvars; rinv=rinv; ractions=acts;
	   res_locals = locals; res_body = body; res_loc=loc}
    | Pdec_indpred (pid,param,body,loc) ->
	let param = List.map (fst >> create_ident) param in
	let body = prop_to_ip_body body loc in
	let node = List.length param = 3 && (match body with
	  | [(x,((_,sl),[]))] when pure_is_true x ->
             begin match spat_fold (fun x r -> x::r) sl [] with
	       | [Csp_node(_,Ce_ident id,_)] -> id = List.hd param
	       | _ -> false
	     end
          | _ -> false) in
	begin try
	  add_ip_decl {ip_id=pid;ip_args=param;ip_body=body;ip_node=node;ip_loc=loc}
	with Indpred_error (s,loc) -> error (s,loc) end
  in 
  List.iter go1 iteml;
  List.iter go2 iteml;
  (* Type checking *)
  let check_ty s (ty,loc) = match ty with
    | Pty_null -> assert false
    | Pty_base _ -> ()
    | Pty_class t ->
	begin 
	  try ignore (lookup_class env t)
	  with Env_not_found -> error("The type of " ^ s ^ " refers to the undeclared class " ^ t ^ ".", loc)
	end
  in
  Hashtbl.iter (fun _ ci -> StringMap.iter (fun x -> check_ty ("field " ^ x)) ci.cla_fields) env.classes;
  IdMap.iter (fun x -> check_ty ("global variable " ^ string_of_ident x)) env.variables;
  List.iter (lookup_resource env >> tycheck_res env) (List.rev env.env_rsrcs);
  List.iter (lookup_function env >> tycheck_fun env) (List.rev env.env_funcs);
  if !Config.verbose1 then print_ip_env !Config.formatter

open Format

let pp_string pp s = fprintf pp "%s" s

let rec pp_listsep pp s1 s2 f = function
  | [] -> fprintf f "%s" s2
  | [x] -> fprintf f "%a" pp x
  | x::l -> fprintf f "%a%s%a" pp x s1 (pp_listsep pp s1 s2) l

let pp_call_item pp (fid,acq) =
  match StringSet.elements acq with
  | [] -> pp_string pp fid
  | l -> fprintf pp "{%a}%a" (pp_listsep pp_string " " "") l pp_string fid

let pp_call_set pp sm =
  match StringMap.fold (fun fid acq l -> (fid,acq)::l) sm [] with
  | [] -> ()
  | call_items -> fprintf pp " CALLS %a" (pp_listsep pp_call_item "," "") call_items

let pp_strset x pp s = match StringSet.elements s with
  | [] -> ()
  | l -> fprintf pp " %s %a" x (pp_listsep pp_string "," "") l

let pp_idset x pp s = match (List.map string_of_ident (IdSet.elements s)) with
  | [] -> ()
  | l -> fprintf pp " %s %a" x (pp_listsep pp_string "," "") l

let print_env pp env =
  let f fid fi = fprintf pp "FUNCTION %s%a%a%a%a%a@." fid
     pp_call_set fi.calls  (pp_strset "IN_PARALLEL_WITH") fi.in_parallel_with (pp_strset "REQUIRES") fi.requires 
     (pp_idset "MODIFIES") fi.modif (pp_idset "READS") fi.vars in
  Hashtbl.iter f env.functions

(* star resource invariants *)
let desugar_rsrc_initzers iteml env =
  let rsrc_invs =
    List.fold_right
      (fun item invs -> match item with
	 | Pdec_resource(rid,prots,Some inv,([],[]),actions,loc) -> (
	     let can_inv = prop_to_can_prop inv loc in
	       can_prop_star can_inv invs)
	 | _ -> invs)
      iteml
      can_prop_empty
  (* if absent, extend iteml with an init function consisting of the
     resource initializers *)
  in (rsrc_invs,
      if (List.exists (function Pdec_fun("init",_,_,_,_,_) -> true | _ -> false) iteml) 
	|| not (List.exists (function Pdec_resource _ -> true | _ -> false) iteml)
      then iteml
      else (Pdec_fun("init", "void",
	 	     ([],[]),
	 	     (Some a_prop_empty, Some a_prop_empty),
	 	     ([],[]),
	 	     Location.none)
	    :: iteml))

let compose_init_and_main rsrc_invs env =
  let init_post =
    try 
      let init_fun = lookup_function env "init" in
      Hashtbl.replace env.functions "init" {init_fun with post = (can_prop_star init_fun.post rsrc_invs)};
      init_fun.post
    with Env_not_found -> can_prop_empty in
  try
    let main_fun = lookup_function env "main" in
    env_add_fun env "compose_init_main" 
      {fn_resty=Pty_base "void";
       pre= init_post;
       body= [];
       post= main_fun.pre;
       param= [],[];
       locals= [];
       bound= IdSet.empty;
       calls= StringMap.empty;
       in_parallel_with= StringSet.empty;
       requires= StringSet.empty;
       writes_heap = false;
       modif= IdSet.empty;
       vars= IdSet.empty;
       fn_loc= Location.none}
  with Env_not_found -> (if !Config.verbose2 then (fprintf !Config.formatter "No main() found.@."))

let check_variable_conditions env =
  let check_main_req_empty () =
    try (* check that main requires no resources *)
      let main_fun = lookup_function env "main" in
      if not (StringSet.is_empty main_fun.requires) then 
	error ("Function main requires resource " ^ StringSet.choose main_fun.requires
		^ " in order to execute.", main_fun.fn_loc)
    with Env_not_found -> () in
  let check_init_sequential () =
    try (* check that init is sequential *)
      let init_fun = lookup_function env "init" in
      Parsetree.iter
	(fun loc -> function
	   | Pstm_withres _ | Pstm_action _ | Pstm_parblock _ ->
	       error ("The body of init must be sequential.", loc)
	   | _ -> ()) init_fun.body;
      if not (StringMap.is_empty init_fun.calls) then
	error ("Function init cannot call other functions.", init_fun.fn_loc)
    with Env_not_found -> () in
  let rec check_par_fcalls loc = function
  | Pstm_parblock par_calls ->
      let get_data (reso,fid,args) =
	let fi = lookup_function env fid in
	let mvr = mvr_body env [mk_fcall_stm loc (reso,fid,args)] in
	let fv = fv_norm_can_prop fi.pre ++ fv_norm_can_prop fi.post in
	(fid,mvr,fv) in
      let check (fid,mvr,fv) (fid',mvr',fv') =
	 empty_or_error (IdSet.inter mvr.md fv')
	   ("Function " ^ fid ^ " modifies ",
	    " which appears in the spec of " ^ fid' ^ ".", loc);
         empty_or_error (IdSet.inter mvr.md mvr'.vr)
	   ("RACE: Function " ^ fid ^ " modifies variable ", 
	    " and function " ^ fid' ^ " accesses it.", loc) in
      let rec go xl yl = match yl with
	| [] -> ()
	| y::yl -> 
	    List.iter (check y) xl;
	    List.iter (check y) yl;
	    go (y::xl) yl in
      go [] (List.map get_data par_calls)
  | _ -> ()
  in 
  check_main_req_empty ();
  check_init_sequential ();
  Hashtbl.iter (fun _ fi -> Parsetree.iter check_par_fcalls fi.body) env.functions

let create_env iteml =
  let env = initial_env () in
  let rsrc_invs, iteml = desugar_rsrc_initzers iteml env in
  process_declarations iteml env;
  compose_init_and_main rsrc_invs env;
  calc_call_graph env;
  calc_mvr env;
  if !Config.verbose1 then print_env !Config.formatter env;
  check_variable_conditions env;
  env

type lv_info = { lv_next: IdSet.t; lv_break: IdSet.t; lv_cont: IdSet.t; lv_return: IdSet.t }

(** [mark_live_vars c lv_post] annotates the live variables before each program point *)
let mark_live_vars c lv_post = 
  let lv_post = {lv_next=lv_post; lv_break=IdSet.empty; lv_cont=IdSet.empty; lv_return=lv_post} in
  let changed = ref false in
  let (+) x y = IdSet.add y x in
  let (-) x y = IdSet.remove y x in
  let rec go c0 lv =
    let set lv_add =
      let lv_add = lv_add ++ (IdSet.singleton tid_var) in
      if IdSet.for_all (fun e -> IdSet.mem e c0.can_st_lv) lv_add 
      then ()
      else (changed := true; c0.can_st_lv <- c0.can_st_lv ++ lv_add); {lv with lv_next = c0.can_st_lv} in
    match c0.can_st_desc with
      | Cst_nondet(c1,c2)        -> set((go_cmd c1 lv).lv_next ++ (go_cmd c2 lv).lv_next)
      | Cst_assign(id,e)         -> set(lv.lv_next - id ++ fv_can_exp e)
      | Cst_fldlookup(id,e,s)    -> set(lv.lv_next + id ++ fv_can_exp e)
      | Cst_fldassign(e1,s,e2)   -> set(lv.lv_next ++ fv_can_exp e1 ++ fv_can_exp e2)
      | Cst_new id               -> set(lv.lv_next - id)
      | Cst_dispose e            -> set(lv.lv_next ++ fv_can_exp e)
      | Cst_pfcall par_calls     ->
	  set(List.fold_left
		 (fun lv (_,_,vl,el) -> 
		    List.fold_left (fun s x -> fv_can_exp x ++ s) 
		      (List.fold_left (+) lv vl) el)
			 (List.fold_left (fun lv (id,_,_,_) -> lv - id) lv.lv_next par_calls) par_calls)
      | Cst_fcall2(modf,cp,cq,s) -> set(lv.lv_next -- modf ++ fv_can_prop cp)
      | Cst_assume (cp)          -> set(lv.lv_next ++ fv_can_prop cp)
      | Cst_region(c1,_,_,_)     -> set((go_cmd c1 lv).lv_next)
      | Cst_as_action(c1,_,cr,cp,cq,_)-> 
	  let lv' = 
	    let fv = fv_can_prop cq in
            { lv_next = lv.lv_next ++ fv; lv_break = lv.lv_break ++ fv;
	      lv_cont = lv.lv_cont ++ fv; lv_return = lv.lv_return ++ fv } in
	  set((go_cmd c1 lv').lv_next ++ fv_can_prop cp ++ fv_can_prop cr)
      | Cst_loop (c1,cpo)       ->
	  let fv = match cpo with None -> IdSet.empty | Some cp -> fv_can_prop cp in
	  let lv' =
	    let x = c0.can_st_lv ++ fv in 
            { lv_next = x; lv_break = lv.lv_next;
	      lv_cont = x; lv_return = lv.lv_return } in
	  set((go_cmd c1 lv').lv_next ++ fv)
      | Cst_goto (Cgo_break)     -> set (lv.lv_break)
      | Cst_goto (Cgo_continue)  -> set (lv.lv_cont)
      | Cst_goto (Cgo_return)    -> set (lv.lv_return)
  and go_cmd c cfg = List.fold_right go c cfg in
  let rec fix () = 
    changed := false;
    let _ = go_cmd c lv_post in
    if !changed then fix () else () in
  fix ()

let vcgen_stm env in_parallel body =
  let mkst desc loc =
    { can_st_desc = desc;
      can_st_lv = IdSet.empty;
      can_st_loc = loc; } in
  let rec f {pstm_desc = d; pstm_loc = loc} = match d with
    | Pstm_exp (e) ->
	let (stmts,_) = p_exp_to_can_exp_se e [] in
	stmts
    | Pstm_assign (i, pe) ->
	(match pe with 
	  | {pexp_desc = Pexp_fld (e,c); pexp_loc = l} ->
	      let (stmts,ce) = p_exp_to_can_exp_se e [] in
	      stmts @ [mkst (Cst_fldlookup (create_ident i, ce, c)) loc]
	  | {pexp_desc = Pexp_new _; pexp_loc = _} -> [mkst (Cst_new (create_ident i)) loc] 
	  | _ -> 
	      let (stmts,ce) = p_exp_to_can_exp_se pe [] in
	      stmts @ [mkst (Cst_assign (create_ident i, ce)) loc])
    | Pstm_fldassign (e1, t, e2) ->
	let (stmts1,ce1) = p_exp_to_can_exp_se e1 [] in
	let (stmts2,ce2) = p_exp_to_can_exp_se e2 [] in
	stmts1 @ stmts2 @ [mkst (Cst_fldassign (ce1,t,ce2)) loc]
    | Pstm_dispose e ->
	let (stmts,ce) = p_exp_to_can_exp_se e [] in
	stmts @[mkst (Cst_dispose ce) loc]
    | Pstm_block sl ->
	List.flatten (List.map f sl)
    | Pstm_if (e, s1, s2) ->
	let (stmts1, stmts2) = match e with 
	  | None -> ([],[]) 
	  | Some e -> p_exp_to_can_prop_se e in
	let c1 = stmts1 @ f s1 in
	let c2 = stmts2 @ f s2 in
        [mkst (Cst_nondet(c1,c2)) loc]
    | Pstm_while (invo, e, s) ->
        let (cp_true, cp_false) = match e with None -> (can_prop_empty,can_prop_empty) | Some e -> 
	  try 
	    p_exp_to_can_prop e
	  with Undef_non_pure_expression -> error ("While loop has a non-pure guard.", loc) in
	let invo = match invo with None -> None | Some inv -> Some (prop_to_can_prop inv loc) in 
	let s1 = mkst (Cst_assume cp_true) loc in
	let s2 = mkst (Cst_assume cp_false) loc in
	let s0 = mkst (Cst_nondet ([s1], [s2; mkst (Cst_goto Cgo_break) loc])) loc in
	[mkst (Cst_loop(s0 :: f s, invo)) loc] 
    | Pstm_return (Some e) ->
	f {pstm_desc = Pstm_assign("Result",e); pstm_loc=loc} @ [mkst (Cst_goto Cgo_return) loc]
    | Pstm_return None -> [mkst (Cst_goto Cgo_return) loc]
    | Pstm_break -> [mkst (Cst_goto Cgo_break) loc]
    | Pstm_continue -> [mkst (Cst_goto Cgo_continue) loc]
    | Pstm_withres (rid, e, s,actid,argl) ->
	let mk_rely aid ai =
	  let f (name,_) =
	    let id = create_ident name in (id, Ce_ident (gensym_garb (into_existential id))) in
          let xel = List.map f ai.act_param in
	  let xel'= List.map (fun id -> (id, Ce_ident (gensym_garb id))) (IdSet.elements ai.act_locals) in
	  (*let sub = mk_subst ((create_ident "TID",Ce_num 102)::xel@xel') in
	    (aid, sub_can_prop sub ai.act_pre, sub_can_prop sub ai.act_post)*)
	  let otid = Ce_ident (gensym_garb (create_ident "_otherTID")) in
	  let sub = mk_subst ((mytid_var,Ce_ident tid_var)::(tid_var,otid)::xel@xel') in
          let pl = pure_conj_one (mk_NEQ(otid,Ce_num 0)) (pure_one (mk_NEQ(otid,Ce_ident tid_var))) in
          (aid, map_prop sub ai.act_ctx, and_cprop_pure (map_prop sub ai.act_pre) pl, map_prop sub ai.act_post) 
        in
	let modified_in_parallel = union_map (function fid -> (lookup_function env fid).modif) (StringSet.elements in_parallel) in
	let (stmts,_) = p_exp_to_can_prop_se e in
	let ri = lookup_resource env rid in
	let modif_body = (mvr_body env [s]).md in
	let modif_on_exit = ri.owned ++ (IdSet.inter ri.rvars modified_in_parallel) in
	let rely = StringMap.fold (fun xid x y -> mk_rely xid x :: y) ri.ractions [] in 
        if actid="" then (* no annotation *) 
          match ri.rinv with
	  | Some inv when rely = [] -> (* a traditional region *)
	    let rely = [("resource invariant",can_prop_empty,inv,inv)] in
	    let body = stmts @ [mkst (Cst_as_action(f s,rid,can_prop_empty,inv,inv,modif_body)) loc] in
            [mkst(Cst_region(body,rid,modif_on_exit,rely)) loc]
	  | _ -> (* a read-only region *)
            [mkst(Cst_region(stmts @ f s, rid, modif_on_exit,rely)) loc]
	else 
	  let ai = try lookup_action ri actid 
	    with Env_not_found -> error ("Action " ^ actid ^ " of resource " ^ rid ^ " is not defined.", loc) in
	  let el = List.map p_exp_to_can_exp argl in
	  let xel = try List.combine (List.map (fst >> create_ident) ai.act_param) el 
            with Invalid_argument _ -> argnum_error ("Action annotation","",ai.act_param,el,loc) in
	  let sub = mk_subst ((mytid_var,Ce_ident tid_var)::xel) in
	  let act_ctx = map_prop sub ai.act_ctx in
	  let act_pre = map_prop sub ai.act_pre in
	  let act_post =  map_prop sub ai.act_post in
	  let body = stmts @ [mkst (Cst_as_action(f s,rid,act_ctx,act_pre,act_post,modif_body)) loc] in          
          [mkst(Cst_region(body,rid,modif_on_exit,rely)) loc]
    | Pstm_action (s,rid,actid,argl) ->
	let modif_body = (mvr_body env [s]).md in
	let ri = lookup_resource env rid in
	let ai = try lookup_action ri actid 
	  with Env_not_found -> error ("Action " ^ actid ^ " of resource " ^ rid ^ " is not defined.", loc) in
	let el = List.map p_exp_to_can_exp argl in
	let xel = try List.combine (List.map (fst >> create_ident) ai.act_param) el 
          with Invalid_argument _ -> argnum_error ("Action annotation","",ai.act_param,el,loc) in
	let sub = mk_subst ((mytid_var,Ce_ident tid_var)::xel) in
	let act_ctx = map_prop sub ai.act_ctx in
	let act_pre = map_prop sub ai.act_pre in
	let act_post =  map_prop sub ai.act_post in
	[mkst (Cst_as_action(f s,rid,act_ctx,act_pre,act_post,modif_body)) loc]          
    | Pstm_parblock par_calls ->
	let go (reso,fid,(vl,el)) =
	  let id = match reso with Some r -> create_ident r | None -> tempid () in
	  let vl = List.map (fst >> create_ident) vl in
	  let el = try List.map p_exp_to_can_exp el
	    with Undef_non_pure_expression -> error ("Parallel function call has a non-pure argument.", loc) in
	  (id,fid,vl,el) in
        [mkst (Cst_pfcall (List.map go par_calls)) loc]
  in
  List.flatten (List.map f body)

let vcgen_fun env res_lv fid =
  let fi = lookup_function env fid in
  let exfv = ex_fv_can_prop fi.pre in
  let sub = existentials_rename_tonormal_sub exfv in
  let pre = map_prop sub fi.pre in
  let pre = and_cprop_pure pre (pure_one ( mk_NEQ(Ce_num 0,Ce_ident tid_var))) in
  let post = map_prop sub fi.post in
  let c = vcgen_stm env fi.in_parallel_with fi.body in
  mark_live_vars c (fv_can_prop post ++ res_lv);
  (fid,(pre,c,post,fi.fn_loc))


(* Check invariant is stable under interference *)
let vcgen_res env rid =
  let mk_rely aid ai =
    let f (name,_) =
      let id = create_ident name in (id, Ce_ident (gensym_garb (into_existential id))) in
    let xel = List.map f ai.act_param in
    let xel'= List.map (fun id -> (id, Ce_ident (gensym_garb id))) (IdSet.elements ai.act_locals) in
    let otid = Ce_ident (gensym_garb (create_ident "_otherTID")) in
    let sub = mk_subst ((mytid_var,Ce_ident tid_var)::(tid_var,otid)::xel@xel') in
    let pl = pure_one (mk_NEQ(otid,Ce_num 0)) in
    (aid, map_prop sub ai.act_ctx, and_cprop_pure (map_prop sub ai.act_pre) pl, map_prop sub ai.act_post)
  in
  let ri = lookup_resource env rid in
  let rely = StringMap.fold (fun xid x y -> mk_rely xid x :: y) ri.ractions [] in
  match ri.rinv with
    | None ->
	let c = vcgen_stm env StringSet.empty ri.res_body in
  	mark_live_vars c ri.rvars;
	(rid, Cri_code (ri.rvars,c,rely,ri.res_loc))
    | Some inv -> (rid, Cri_inv (inv,rely,ri.res_loc))


let make_fun_htable env =
  let ht = Hashtbl.create 32 in
  let go fi = 
    {fun_param = (List.map (fun(x,_,_)->x) (fst fi.param), List.map (fun(x,_,_)->x) (snd fi.param));
     fun_modif = fi.modif;
     fun_pre = fi.pre;
     fun_post = fi.post;
     fun_loc = fi.fn_loc}  in
  Hashtbl.iter (fun x fi -> Hashtbl.add ht x (go fi)) env.functions;
  ht

let vcgen iteml =
  Inductive_preds.init_ip_env ();
  error_handler (fun () ->
    let env = create_env iteml in
    let entl1 = List.rev_map (vcgen_res env) env.env_rsrcs in
    let res_lv = Hashtbl.fold (fun _ x r -> x.rvars ++ r)
		 env.resources IdSet.empty in
    let entl2 = List.rev_map (vcgen_fun env res_lv) env.env_funcs in
    let  _ = try check_ip_uses () with Indpred_error (s,loc) -> error (s,loc) in
    let ht = make_fun_htable env in
    (ht, entl1, entl2))
